from kubernetes import client, config

from tempo.conf import settings

RCloneConf = """
[s3]
type = s3
provider = minio
env_auth = false
access_key_id = minioadmin
secret_access_key = minioadmin
endpoint = http://{MINIO_IP}:9000
"""


def create_minio_rclone(path: str):
    config.load_kube_config()
    api_instance = client.CoreV1Api()
    res = api_instance.list_namespaced_service("minio-system", field_selector="metadata.name=minio")
    minio_ip = res.items[0].status.load_balancer.ingress[0].ip
    rclone_conf = RCloneConf.replace("{MINIO_IP}", minio_ip)
    with open(path, "w") as f:
        f.write(rclone_conf)
    settings.rclone_cfg = path



import os
import time
from enum import Enum
from http import cookies
from typing import Any, Dict

from pydantic import BaseModel
from seldon_deploy_sdk import ApiClient, Configuration, EnvironmentApi, SeldonDeploymentsApi
from seldon_deploy_sdk.auth import OIDCAuthenticator, SessionAuthenticator
from seldon_deploy_sdk.models.object_meta import ObjectMeta
from seldon_deploy_sdk.models.predictive_unit import PredictiveUnit
from seldon_deploy_sdk.models.predictor_spec import PredictorSpec
from seldon_deploy_sdk.models.seldon_deployment import SeldonDeployment
from seldon_deploy_sdk.models.seldon_deployment_spec import SeldonDeploymentSpec

from tempo.seldon.endpoint import Endpoint
from tempo.seldon.k8s import SeldonKubernetesRuntime
from tempo.seldon.specs import KubernetesSpec
from tempo.serve.metadata import ModelDetails
from tempo.serve.remote import Remote
from tempo.serve.runtime import ModelSpec, Runtime


class SeldonDeployAuthType(Enum):
    session_cookie = "session_cookie"
    oidc = "oidc"


class SeldonDeployConfig(BaseModel):

    host: str
    user: str
    password: str
    auth_type: SeldonDeployAuthType = SeldonDeployAuthType.session_cookie
    oidc_client_id: str
    oidc_server: str
    verify_ssl: bool = True


class SeldonDeployRuntime(Runtime, Remote):
    def __init__(self):
        self.api_client = None

    def authenticate(self, settings: SeldonDeployConfig):
        config = Configuration()
        config.host = settings.host
        config.verify_ssl = settings.verify_ssl
        if settings.auth_type == SeldonDeployAuthType.session_cookie:
            auth = SessionAuthenticator(config)
            # TODO can we use just cookie_str in API client
            cookie_str = auth.authenticate(settings.user, settings.password)
            ckies: cookies.SimpleCookie = cookies.SimpleCookie()
            ckies.load(cookie_str)
            token = ckies["authservice_session"].value
            cookie_str = f"authservice_session={token}"
            api_client = ApiClient(config, cookie=cookie_str)
            self.api_client = api_client
        elif settings.auth_type == SeldonDeployAuthType.oidc:
            if not settings.verify_ssl:
                os.environ["CURL_CA_BUNDLE"] = ""
            config.oidc_client_id = settings.oidc_client_id
            config.oidc_server = settings.oidc_server
            auth = OIDCAuthenticator(config)
            config.access_token = auth.authenticate(settings.user, settings.password)
            api_client = ApiClient(config)
            self.api_client = api_client
        else:
            raise ValueError(f"Unknown auth type {settings.auth_type}")

    def _create_api_client(self):
        if self.api_client is not None:
            return self.api_client
        else:
            # TODO: auto authenticate from environment
            pass

    def deploy_spec(self, model_spec: ModelSpec):
        self._create_api_client()
        env_api = EnvironmentApi(self.api_client)
        user = env_api.read_user()

        print(user)

        # TODO: Use KubernetesSpec().spec
        sd = SeldonDeployment(
            kind="SeldonDeployment",
            api_version="machinelearning.seldon.io/v1",
            metadata=ObjectMeta(
                name=model_spec.model_details.name, namespace=model_spec.runtime_options.k8s_options.namespace
            ),
            spec=SeldonDeploymentSpec(
                predictors=[
                    PredictorSpec(
                        graph=PredictiveUnit(
                            implementation=KubernetesSpec.Implementations[model_spec.model_details.platform],
                            model_uri=model_spec.model_details.uri,
                            name="model",
                        ),
                        name="default",
                        replicas=model_spec.runtime_options.k8s_options.replicas,
                    )
                ]
            ),
        )

        dep_instance = SeldonDeploymentsApi(self.api_client)
        dep_instance.create_seldon_deployment(model_spec.runtime_options.k8s_options.namespace, sd)

    def wait_ready_spec(self, model_spec: ModelSpec, timeout_secs=None) -> bool:
        self._create_api_client()
        ready = False
        t0 = time.time()
        dep_instance = SeldonDeploymentsApi(self.api_client)
        while not ready:
            sdep: SeldonDeployment = dep_instance.read_seldon_deployment(
                model_spec.model_details.name, model_spec.runtime_options.k8s_options.namespace
            )
            sdep_dict = sdep.to_dict()
            ready = sdep_dict["status"]["state"] == "Available"
            if timeout_secs is not None:
                t1 = time.time()
                if t1 - t0 > timeout_secs:
                    return ready
        return ready

    def undeploy_spec(self, model_spec: ModelSpec):
        self._create_api_client()
        dep_instance = SeldonDeploymentsApi(self.api_client)
        dep_instance.delete_seldon_deployment(
            model_spec.model_details.name, model_spec.runtime_options.k8s_options.namespace, _preload_content=False
        )

    def get_endpoint_spec(self, model_spec: ModelSpec):
        endpoint = Endpoint()
        return endpoint.get_url(model_spec)

    def get_headers(self, model_details: ModelDetails) -> Dict[str, str]:
        return {}

    def remote(self, model_spec: ModelSpec, *args, **kwargs) -> Any:
        self._create_api_client()
        srt = SeldonKubernetesRuntime()
        srt.remote(model_spec, *args, **kwargs)

    def to_k8s_yaml_spec(self, model_spec: ModelSpec) -> str:
        srt = SeldonKubernetesRuntime()
        return srt.to_k8s_yaml_spec(model_spec)



from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.seldon.protocol import SeldonProtocol
from tempo.seldon.specs import KubernetesSpec, _V2ContainerFactory, get_container_spec
from tempo.serve.metadata import KubernetesOptions, ModelDataArgs, ModelDetails, ModelFramework, RuntimeOptions
from tempo.serve.model import Model
from tempo.serve.runtime import ModelSpec


def test_kubernetes_spec(sklearn_model: Model):
    k8s_object = KubernetesSpec(sklearn_model.model_spec)

    expected = {
        "apiVersion": "machinelearning.seldon.io/v1",
        "kind": "SeldonDeployment",
        "metadata": {
            "name": sklearn_model.details.name,
            "namespace": sklearn_model.model_spec.runtime_options.k8s_options.namespace,
        },
        "spec": {
            "protocol": "seldon",
            "predictors": [
                {
                    "graph": {
                        "modelUri": sklearn_model.details.uri,
                        "name": "test-iris-sklearn",
                        "type": "MODEL",
                        "implementation": KubernetesSpec.Implementations[sklearn_model.details.platform],
                    },
                    "name": "default",
                    "replicas": sklearn_model.model_spec.runtime_options.k8s_options.replicas,
                }
            ],
        },
    }

    assert k8s_object.spec == expected


def test_kubernetes_spec_pipeline():
    details = ModelDetails(
        name="inference-pipeline",
        platform=ModelFramework.TempoPipeline,
        uri="gs://seldon/tempo",
        local_folder="",
        inputs=ModelDataArgs(args=[]),
        outputs=ModelDataArgs(args=[]),
    )
    options = KubernetesOptions(namespace="production", replicas=1)
    protocol = KFServingV2Protocol()
    runtime_options = RuntimeOptions(k8s_options=options)
    model_spec = ModelSpec(model_details=details, protocol=protocol, runtime_options=runtime_options)
    k8s_object = KubernetesSpec(model_spec)

    container_spec = _V2ContainerFactory.get_container_spec(details, runtime_options)
    container_env = [{"name": name, "value": value} for name, value in container_spec["environment"].items()]

    expected = {
        "apiVersion": "machinelearning.seldon.io/v1",
        "kind": "SeldonDeployment",
        "metadata": {"name": details.name, "namespace": options.namespace},
        "spec": {
            "protocol": "kfserving",
            "predictors": [
                {
                    "componentSpecs": [
                        {
                            "spec": {
                                "containers": [
                                    {
                                        "name": "inference-pipeline",
                                        "image": container_spec["image"],
                                        "env": container_env,
                                        "args": [],
                                    }
                                ]
                            }
                        }
                    ],
                    "graph": {
                        "modelUri": details.uri,
                        "name": "inference-pipeline",
                        "type": "MODEL",
                        "implementation": "TRITON_SERVER",
                        "serviceAccountName": "tempo-pipeline",
                    },
                    "name": "default",
                    "replicas": options.replicas,
                }
            ],
        },
    }

    assert k8s_object.spec == expected


def test_tensorflow_spec():
    md = ModelDetails(
        name="test",
        local_folder="",
        uri="",
        platform=ModelFramework.Tensorflow,
        inputs=ModelDataArgs(args=[]),
        outputs=ModelDataArgs(args=[]),
    )
    protocol = SeldonProtocol()
    options = KubernetesOptions(namespace="production", replicas=1)
    runtime_options = RuntimeOptions(k8s_options=options)
    model_spec = ModelSpec(model_details=md, protocol=protocol, runtime_options=runtime_options)
    spec = get_container_spec(model_spec)
    assert "image" in spec
    assert "command" in spec



import time

import docker
import numpy as np
import pytest

from tempo.seldon.docker import SeldonDockerRuntime
from tempo.serve.model import Model


def test_deploy_docker(sklearn_model: Model, runtime: SeldonDockerRuntime):

    time.sleep(2)

    container = runtime._get_container(sklearn_model.model_spec)
    assert container.status == "running"


@pytest.mark.parametrize(
    "x_input",
    [[[1, 2, 3, 4]], np.array([[1, 2, 3, 4]]), {"data": {"ndarray": [[1, 2, 3, 4]]}}],
)
def test_sklearn_docker(sklearn_model: Model, x_input):
    time.sleep(2)

    y_pred = sklearn_model(x_input)

    np.testing.assert_allclose(y_pred, [[0, 0, 0.99]], atol=1e-2)


def test_undeploy_docker(sklearn_model: Model, runtime: SeldonDockerRuntime):
    time.sleep(2)

    runtime.undeploy(sklearn_model)

    with pytest.raises(docker.errors.NotFound):
        runtime._get_container(sklearn_model.model_spec)



import pytest

from tempo.serve.metadata import RuntimeOptions


@pytest.mark.parametrize(
    "runtime, replicas",
    [
        ({"k8s_options": {"replicas": 2}}, 2),
        ({}, 1),
    ],
)
def test_runtime_options(runtime, replicas):
    r = RuntimeOptions(**runtime)
    assert r.k8s_options.replicas == replicas



from types import SimpleNamespace
from typing import Any, Callable, Optional

from tempo.serve.base import BaseModel
from tempo.serve.constants import ModelDataType
from tempo.serve.metadata import ModelFramework, RuntimeOptions
from tempo.serve.model import Model
from tempo.serve.protocol import Protocol
from tempo.serve.runtime import Runtime


class PipelineModels(SimpleNamespace):
    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def remote_copy(self):
        output = {}
        for name, model in self.items():
            output[name] = Model(
                name=model.get_tempo().details.name,
                local_folder=model.get_tempo().details.local_folder,
                uri=model.get_tempo().details.uri,
                platform=model.get_tempo().details.platform,
                inputs=model.get_tempo().details.inputs,
                outputs=model.get_tempo().details.outputs,
                protocol=model.get_tempo().protocol,
                runtime_options=model.get_tempo().model_spec.runtime_options,
            )
        return PipelineModels(**output)


class Pipeline(BaseModel):
    def __init__(
        self,
        name: str,
        pipeline_func: Callable[[Any], Any] = None,
        protocol: Optional[Protocol] = None,
        models: PipelineModels = None,
        local_folder: str = None,
        uri: str = None,
        inputs: ModelDataType = None,
        outputs: ModelDataType = None,
        conda_env: str = None,
        runtime_options: RuntimeOptions = RuntimeOptions(),
    ):
        super().__init__(
            name=name,
            # TODO: Should we unify names?
            user_func=pipeline_func,
            local_folder=local_folder,
            uri=uri,
            platform=ModelFramework.TempoPipeline,
            inputs=inputs,
            outputs=outputs,
            conda_env=conda_env,
            protocol=protocol,
            runtime_options=runtime_options,
        )

        if models is None:
            models = PipelineModels()

        self.models = models

    def save(self, save_env=True):
        # Backup models and create lite copy so Pipeline doesn not inherit dependencies
        models_backup = self.models
        self.models = self.models.remote_copy()
        super().save(save_env=save_env)

        # Restore backed up models
        self.models = models_backup

    def set_remote(self, val: bool):
        for model in self.models.values():
            model.get_tempo().set_remote(val)

    def set_runtime_options_override(self, runtime_options: RuntimeOptions):
        for model in self.models.values():
            model.get_tempo().set_runtime_options_override(runtime_options)
        super().set_runtime_options_override(runtime_options)

    def deploy_models(self, runtime: Runtime):
        for model in self.models.values():
            model.get_tempo().deploy(runtime)

    def deploy(self, runtime: Runtime):
        self.deploy_models(runtime)
        super().deploy(runtime)

    def wait_ready(self, runtime: Runtime, timeout_secs: int = None) -> bool:
        super().wait_ready(runtime, timeout_secs=timeout_secs)
        for model in self.models.values():
            if not model.get_tempo().wait_ready(runtime, timeout_secs=timeout_secs):
                return False
        return True

    def undeploy_models(self, runtime: Runtime):
        for model in self.models.values():
            model.get_tempo().undeploy(runtime)

    def undeploy(self, runtime: Runtime):
        """
        Undeploy all models and pipeline.
        """
        super().undeploy(runtime)
        self.undeploy_models(runtime)

    def to_k8s_yaml(self, runtime: Runtime) -> str:
        yamls = super().to_k8s_yaml(runtime)
        yamls += "\n---\n"
        for model in self.models.values():
            y = model.get_tempo().to_k8s_yaml(runtime)
            yamls += y
            yamls += "\n---\n"
        return yamls



from typing import Tuple

from tempo.serve.utils import pipeline, predictmethod


def test_class_func_class():
    @pipeline(
        name="classifier",
        models=[],
    )
    class MyPipeline:
        @predictmethod
        def predict(self, X: str) -> str:
            return X

    x = MyPipeline()

    r = x.predict("hello")
    assert r == "hello"
    r = x("hello")
    assert r == "hello"


def test_class_func():
    @pipeline(
        name="classifier",
        models=[],
    )
    def predict(X: str) -> str:
        return X

    r = predict("hello")
    assert r == "hello"


def test_clear_state_func():
    @pipeline(
        name="classifier",
        models=[],
    )
    class MyPipeline:
        def __init__(self):
            self.cleared = False

        @predictmethod
        def predict(self, X: str) -> str:
            return X

    x = MyPipeline()

    @pipeline(
        name="classifier",
        models=[x],
    )
    class MyPipeline2:
        def __init__(self):
            self.cleared = False

        @predictmethod
        def predict(self, X: str) -> str:
            return x(X=X)

    y = MyPipeline2()

    y(X="hello")


def test_class_two_outputs():
    @pipeline(
        name="classifier",
        models=[],
    )
    class MyPipeline:
        @predictmethod
        def predict(self, X: str) -> Tuple[str, str]:
            return X, X

    x = MyPipeline()

    r1, r2 = x.predict("hello")
    assert r1 == "hello"
    assert r2 == "hello"



import json
import os

from mlserver import MLModel
from mlserver.types import InferenceRequest, InferenceResponse
from mlserver.utils import get_model_uri

from tempo.serve.constants import ENV_TEMPO_RUNTIME_OPTIONS
from tempo.serve.loader import load
from tempo.serve.metadata import RuntimeOptions


class InferenceRuntime(MLModel):
    async def load(self) -> bool:
        pipeline_uri = await get_model_uri(self._settings)
        self._pipeline = load(pipeline_uri)
        rt_options_str = os.getenv(ENV_TEMPO_RUNTIME_OPTIONS)
        if rt_options_str:
            rt_options = RuntimeOptions(**json.loads(rt_options_str))
            self._pipeline.set_runtime_options_override(rt_options)
        self.ready = True
        return self.ready

    async def predict(self, payload: InferenceRequest) -> InferenceResponse:
        prediction = self._pipeline.request(payload.dict())
        return InferenceResponse(**prediction)



import os
from pathlib import Path

import numpy as np
import pytest
import yaml

from tempo import Model, ModelFramework, Pipeline, pipeline, predictmethod
from tempo.kfserving.protocol import KFServingV1Protocol, KFServingV2Protocol
from tempo.seldon.protocol import SeldonProtocol
from tempo.serve.constants import MLServerEnvDeps
from tempo.serve.metadata import KubernetesOptions, RuntimeOptions
from tempo.serve.pipeline import PipelineModels
from tempo.serve.utils import model

TESTS_PATH = os.path.dirname(__file__)
TESTDATA_PATH = os.path.join(TESTS_PATH, "testdata")
PIPELINE_LOCAL_DIR = os.path.join(TESTS_PATH, "artifacts/pipeline")


def pytest_collection_modifyitems(items):
    """
    Add pytest.mark.asyncio marker to every test.
    """
    for item in items:
        item.add_marker("asyncio")


@pytest.fixture
def pipeline_conda_yaml() -> str:
    condaPath = PIPELINE_LOCAL_DIR + "/conda.yaml"
    if not os.path.isfile(condaPath):
        with open(PIPELINE_LOCAL_DIR + "/conda.yaml.tmpl") as f:
            env = yaml.safe_load(f)
            path = Path(TESTS_PATH)
            parent = path.parent.absolute()
            pip_deps = {"pip": []}
            env["dependencies"].append(pip_deps)
            for dep in MLServerEnvDeps:
                pip_deps["pip"].append(dep)
            pip_deps["pip"].append("mlops-tempo @ file://" + str(parent))
            with open(condaPath, "w") as f2:
                yaml.safe_dump(env, f2)
    return condaPath


@pytest.fixture
def sklearn_model() -> Model:
    model_path = os.path.join(TESTDATA_PATH, "sklearn", "iris")
    return Model(
        name="test-iris-sklearn",
        platform=ModelFramework.SKLearn,
        uri="gs://seldon-models/sklearn/iris",
        local_folder=model_path,
        protocol=SeldonProtocol(),
        runtime_options=RuntimeOptions(k8s_options=KubernetesOptions(namespace="production", replicas=1)),
    )


@pytest.fixture
def xgboost_model() -> Model:
    model_path = os.path.join(TESTDATA_PATH, "xgboost", "iris")
    return Model(
        name="test-iris-xgboost",
        platform=ModelFramework.XGBoost,
        uri="gs://seldon-models/xgboost/iris",
        local_folder=model_path,
        protocol=SeldonProtocol(),
    )


@pytest.fixture
def custom_model() -> Model:
    @model(
        name="custom-model",
        protocol=KFServingV2Protocol(),
        platform=ModelFramework.Custom,
    )
    def _custom_model(payload: np.ndarray) -> np.ndarray:
        return payload.sum(keepdims=True)

    return _custom_model


@pytest.fixture
def inference_pipeline(sklearn_model: Model, xgboost_model: Model, pipeline_conda_yaml: str) -> Pipeline:
    @pipeline(
        name="inference-pipeline",
        models=PipelineModels(sklearn=sklearn_model, xgboost=xgboost_model),
        uri="gs://seldon-models/tempo/test_pipeline",
        local_folder=PIPELINE_LOCAL_DIR,
    )
    def _pipeline(payload: np.ndarray) -> np.ndarray:
        res1 = _pipeline.models.sklearn(payload)
        if res1[0][0] > 0.7:
            return res1
        else:
            return _pipeline.models.xgboost(payload)

    return _pipeline


@pytest.fixture
def cifar10_model() -> Model:
    model_path = os.path.join(TESTDATA_PATH, "tfserving", "cifar10", "resnet32")
    return Model(
        name="resnet32",
        platform=ModelFramework.Tensorflow,
        uri="gs://seldon-models/tfserving/cifar10/resnet32",
        local_folder=model_path,
        protocol=KFServingV1Protocol(),
    )


@pytest.fixture
def inference_pipeline_class(sklearn_model: Model, xgboost_model: Model):
    @pipeline(
        name="mypipeline",
        models=PipelineModels(sklearn=sklearn_model, xgboost=xgboost_model),
        local_folder=PIPELINE_LOCAL_DIR,
    )
    class MyClass(object):
        def __init__(self):
            self.counter = 0

        @predictmethod
        def p(self, payload: np.ndarray) -> np.ndarray:
            self.counter += 1
            return payload.sum(keepdims=True)

        def get_counter(self):
            return self.counter

    myc = MyClass()
    return myc



import docker
import numpy as np
import pytest
import yaml

from tempo.seldon.docker import SeldonDockerRuntime
from tempo.serve.constants import MLServerEnvDeps
from tempo.serve.pipeline import Pipeline


def test_conda_yaml(pipeline_conda_yaml):
    print(pipeline_conda_yaml)
    with open(pipeline_conda_yaml) as f:
        env = yaml.safe_load(f)
        for dep in env["dependencies"]:
            if dep == "pip":
                assert dep[0] == MLServerEnvDeps[0]


def test_deploy_pipeline_docker(
    inference_pipeline: Pipeline,
    runtime: SeldonDockerRuntime,
):
    for model in inference_pipeline.models.values():
        container = runtime._get_container(model.model_spec)
        assert container.status == "running"

    pipeline_container = runtime._get_container(inference_pipeline.model_spec)
    assert pipeline_container.status == "running"


@pytest.mark.parametrize(
    "x_input",
    [
        pytest.param(
            [[1, 2, 3, 4]],
            marks=pytest.mark.skip(reason="not supported with KFServingV2Protocol"),
        ),
        np.array([[0.5, 2, 3, 4]]),
        pytest.param(
            {
                "inputs": [
                    {
                        "name": "payload",
                        "shape": [1, 4],
                        "datatype": "INT32",
                        "data": [[0.4, 2, 3, 4]],
                    }
                ]
            },
            marks=pytest.mark.skip(reason="not supported with KFServingV2Protocol"),
        ),
    ],
)
def test_pipeline_remote(inference_pipeline: Pipeline, runtime: SeldonDockerRuntime, x_input):
    y_pred = inference_pipeline.remote(payload=x_input)

    np.testing.assert_allclose(y_pred, [2.0], atol=1e-2)


@pytest.mark.parametrize(
    "x_input, expected",
    [
        (
            {
                "inputs": [
                    {
                        "name": "input0",
                        "datatype": "FP64",
                        "shape": [1, 4],
                        "data": [0.4, 2, 3, 4],
                    }
                ]
            },
            {
                "model_name": "inference-pipeline",
                "outputs": [{"name": "output0", "datatype": "FP64", "shape": [1], "data": [2.0]}],
            },
        )
    ],
)
def test_seldon_pipeline_request_docker(inference_pipeline: Pipeline, x_input, expected):
    y_pred = inference_pipeline.request(x_input)

    assert y_pred == expected


def test_undeploy_pipeline_docker(inference_pipeline: Pipeline, runtime: SeldonDockerRuntime):
    runtime.undeploy(inference_pipeline)

    for model in inference_pipeline.models.values():
        with pytest.raises(docker.errors.NotFound):
            runtime._get_container(model.model_spec)



import pytest

from tempo.seldon.k8s import SeldonKubernetesRuntime
from tempo.seldon.protocol import SeldonProtocol
from tempo.serve.metadata import ModelFramework
from tempo.serve.model import Model


@pytest.mark.parametrize(
    "expected",
    [
        """apiVersion: machinelearning.seldon.io/v1
kind: SeldonDeployment
metadata:
  name: test-iris-sklearn
  namespace: default
spec:
  predictors:
  - graph:
      implementation: SKLEARN_SERVER
      modelUri: gs://seldon-models/sklearn/iris
      name: test-iris-sklearn
      type: MODEL
    name: default
    replicas: 1
  protocol: seldon
"""
    ],
)
def test_seldon_sklearn_model_yaml(expected):
    m = Model(
        name="test-iris-sklearn",
        platform=ModelFramework.SKLearn,
        uri="gs://seldon-models/sklearn/iris",
        protocol=SeldonProtocol(),
        local_folder="",
    )
    runtime = SeldonKubernetesRuntime()
    assert runtime.to_k8s_yaml(m) == expected


@pytest.mark.parametrize(
    "expected",
    [
        """apiVersion: machinelearning.seldon.io/v1
kind: SeldonDeployment
metadata:
  name: test-iris-xgboost
  namespace: default
spec:
  predictors:
  - graph:
      implementation: XGBOOST_SERVER
      modelUri: gs://seldon-models/xgboost/iris
      name: test-iris-xgboost
      type: MODEL
    name: default
    replicas: 1
  protocol: seldon
"""
    ],
)
def test_seldon_xgboost_model_yaml(expected):
    m = Model(
        name="test-iris-xgboost",
        protocol=SeldonProtocol(),
        platform=ModelFramework.XGBoost,
        uri="gs://seldon-models/xgboost/iris",
        local_folder="",
    )
    runtime = SeldonKubernetesRuntime()
    assert runtime.to_k8s_yaml(m) == expected



from __future__ import annotations

import abc
from typing import Any, Dict

import attr

from tempo.serve.metadata import ModelDataArgs, ModelDetails


@attr.s(auto_attribs=True)
class Protocol(abc.ABC):
    @abc.abstractmethod
    def to_protocol_request(self, *args, **kwargs) -> Dict:
        pass

    @abc.abstractmethod
    def to_protocol_response(self, model_details: ModelDetails, *args, **kwargs) -> Dict:
        pass

    @abc.abstractmethod
    def from_protocol_request(self, res: Dict, tys: ModelDataArgs) -> Any:
        pass

    @abc.abstractmethod
    def from_protocol_response(self, res: Dict, tys: ModelDataArgs) -> Any:
        pass

    @abc.abstractmethod
    def get_predict_path(self, model_details: ModelDetails) -> str:
        pass

    @abc.abstractmethod
    def get_status_path(self, model_details: ModelDetails) -> str:
        pass



import base64
import inspect
import json
from typing import Any

from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.serve.constants import ModelDataType
from tempo.serve.metadata import ModelFramework, RuntimeOptions
from tempo.serve.model import Model
from tempo.serve.pipeline import Pipeline, PipelineModels
from tempo.serve.protocol import Protocol


def b64_encode_runtime_options(runtime_options: RuntimeOptions):
    return str(base64.b64encode(json.dumps(runtime_options.dict()).encode("utf8")), "utf-8")


def b64_decode_runtime_options(val: str):
    return RuntimeOptions(**json.loads(base64.b64decode(val.encode("utf-8")).decode("utf-8")))


def pipeline(
    name: str,
    protocol: Protocol = KFServingV2Protocol(),
    local_folder: str = None,
    uri: str = None,
    models: PipelineModels = None,
    inputs: ModelDataType = None,
    outputs: ModelDataType = None,
    conda_env: str = None,
    runtime_options: RuntimeOptions = RuntimeOptions(),
):
    def _pipeline(f):
        if inspect.isclass(f):
            K = f
            func = None

            for a in dir(K):
                if not a.startswith("__") and callable(getattr(K, a)) and hasattr(getattr(K, a), "predict"):
                    func = getattr(K, a)
                    break
            K.pipeline = Pipeline(
                name,
                local_folder=local_folder,
                uri=uri,
                models=models,
                inputs=inputs,
                outputs=outputs,
                pipeline_func=func,
                conda_env=conda_env,
                protocol=protocol,
                runtime_options=runtime_options,
            )
            setattr(K, "request", K.pipeline.request)
            setattr(K, "remote", K.pipeline.remote)
            setattr(K, "get_tempo", K.pipeline.get_tempo)

            orig_init = K.__init__

            # Make copy of original __init__, so we can call it without recursion
            def __init__(self, *args, **kws):
                K.pipeline.set_cls(self)
                orig_init(self, *args, **kws)  # Call the original __init__

            K.__init__ = __init__  # Set the class' __init__ to the new one

            def __call__(self, *args, **kwargs) -> Any:
                return self.pipeline(*args, **kwargs)

            K.__call__ = __call__

            @property
            def models_property(self):
                # This is so we do not store reference to `.models` as part of
                # the K class - needed when saving limited copy of models for remote
                return K.pipeline.models

            K.models = models_property

            return K
        else:
            return Pipeline(
                name,
                local_folder=local_folder,
                uri=uri,
                models=models,
                inputs=inputs,
                outputs=outputs,
                pipeline_func=f,
                conda_env=conda_env,
                protocol=protocol,
                runtime_options=runtime_options,
            )

    return _pipeline


def predictmethod(f):
    f.predict = True
    return f


def model(
    name: str,
    local_folder: str = None,
    uri: str = None,
    platform: ModelFramework = None,
    inputs: ModelDataType = None,
    outputs: ModelDataType = None,
    conda_env: str = None,
    protocol: Protocol = KFServingV2Protocol(),
    runtime_options: RuntimeOptions = RuntimeOptions(),
):
    def _model(f):
        if inspect.isclass(f):
            K = f
            func = None

            for a in dir(K):
                if not a.startswith("__") and callable(getattr(K, a)) and hasattr(getattr(K, a), "predict"):
                    func = getattr(K, a)
                    break

            K.pipeline = Model(
                name,
                protocol=protocol,
                local_folder=local_folder,
                uri=uri,
                platform=platform,
                inputs=inputs,
                outputs=outputs,
                model_func=func,
                conda_env=conda_env,
                runtime_options=runtime_options,
            )

            setattr(K, "request", K.pipeline.request)
            setattr(K, "remote", K.pipeline.remote)
            setattr(K, "get_tempo", K.pipeline.get_tempo)

            orig_init = K.__init__

            # Make copy of original __init__, so we can call it without recursion
            def __init__(self, *args, **kws):
                K.pipeline.set_cls(self)
                orig_init(self, *args, **kws)  # Call the original __init__

            K.__init__ = __init__  # Set the class' __init__ to the new one

            def __call__(self, *args, **kwargs) -> Any:
                return self.pipeline(*args, **kwargs)

            K.__call__ = __call__

            return K
        else:
            return Model(
                name,
                protocol=protocol,
                local_folder=local_folder,
                uri=uri,
                platform=platform,
                inputs=inputs,
                outputs=outputs,
                model_func=f,
                conda_env=conda_env,
                runtime_options=runtime_options,
            )

    return _model



import logging
import os


def _get_logger():
    logger = logging.getLogger("tempo")
    logger.setLevel(logging.INFO)
    return logger


def _get_env():
    if os.environ.get("a", None):
        pass


logger = _get_logger()



import numpy as np
import pytest
from kubernetes import client

from tempo.seldon.k8s import SeldonKubernetesRuntime
from tempo.serve.metadata import RuntimeOptions
from tempo.serve.model import Model
from tempo.serve.pipeline import Pipeline


def test_create_k8s_runtime():
    rto = RuntimeOptions()
    rt = SeldonKubernetesRuntime(rto)
    assert rt.runtime_options.runtime == "tempo.seldon.SeldonKubernetesRuntime"


@pytest.mark.skip(reason="needs k8s cluster")
def test_deploy_k8s(sklearn_model: Model, runtime: SeldonKubernetesRuntime):
    crd_api = client.CustomObjectsApi()
    sdep = crd_api.get_namespaced_custom_object(
        "machinelearning.seldon.io",
        "v1",
        runtime.k8s_options.namespace,
        "seldondeployments",
        sklearn_model.details.name,
    )

    assert sdep["status"]["state"] == "Available"


@pytest.mark.skip(reason="needs k8s cluster")
@pytest.mark.parametrize(
    "x_input",
    [[[1, 2, 3, 4]], np.array([[1, 2, 3, 4]]), {"data": {"ndarray": [[1, 2, 3, 4]]}}],
)
def test_sklearn_k8s(sklearn_model: Model, x_input):
    y_pred = sklearn_model(x_input)

    np.testing.assert_allclose(y_pred, [[0, 0, 0.99]], atol=1e-2)


@pytest.mark.skip(reason="needs k8s cluster")
@pytest.mark.parametrize("x_input", [np.array([[1, 2, 3, 4]])])
def test_pipeline_k8s(inference_pipeline: Pipeline, x_input):
    y_pred = inference_pipeline.remote(payload=x_input)

    np.testing.assert_allclose(y_pred, [2.0], atol=1e-2)



import abc
from typing import Any, Optional

from tempo.serve.metadata import RuntimeOptions
from tempo.serve.runtime import ModelSpec


class Remote(abc.ABC):
    def __init__(self, runtime_options: Optional[RuntimeOptions]):
        self.runtime_options = runtime_options

    @abc.abstractmethod
    def remote(self, model_spec: ModelSpec, *args, **kwargs) -> Any:
        pass



from typing import Dict, List, Optional, Type, Union

SKLEARN_MODEL = "sklearn"
XGBOOST_MODEL = "xgboost"
TENSORFLOW_MODEL = "tensorflow"
PYTORCH_MODEL = "pytorch"
ONNX_MODEL = "onnx"

# PyDantic issue doesn't support raw Tuple : https://github.com/samuelcolvin/pydantic/issues/2132
# ModelDataType = Optional[Union[Type,Tuple,Dict[str,Type]]]
ModelDataType = Optional[Union[Type, List, Dict[str, Type]]]

DefaultModelFilename = "model.pickle"
DefaultRemoteFilename = "remote.pickle"
DefaultEnvFilename = "environment.tar.gz"

# TODO: Update once tempo is published
MLServerEnvDeps = ["mlserver==0.3.1.dev7"]
DefaultCondaFile = "conda.yaml"

ENV_K8S_SERVICE_HOST = "KUBERNETES_SERVICE_HOST"

ENV_TEMPO_RUNTIME_OPTIONS = "TEMPO_RUNTIME_OPTIONS"



import os
from urllib.parse import urlparse

from kubernetes import client, config

from tempo.seldon.specs import DefaultServiceAccountName
from tempo.serve.ingress import create_ingress
from tempo.serve.runtime import ModelSpec
from tempo.utils import logger

ENV_K8S_SERVICE_HOST = "KUBERNETES_SERVICE_HOST"
ISTIO_GATEWAY = "istio"


class Endpoint(object):
    """A Model Endpoint"""

    def __init__(self):
        self.inside_cluster = os.getenv(ENV_K8S_SERVICE_HOST)
        try:
            if self.inside_cluster:
                logger.debug("Loading cluster local config")
                config.load_incluster_config()
            else:
                logger.debug("Loading external kubernetes config")
                config.load_kube_config()
        except Exception:
            logger.warning("Failed to load kubeconfig. Only local mode is possible.")

    def get_service_host(self, model_spec: ModelSpec):
        if self.inside_cluster is not None:
            config.load_incluster_config()
        api_instance = client.CustomObjectsApi()
        api_response = api_instance.get_namespaced_custom_object_status(
            "serving.kubeflow.org",
            "v1beta1",
            model_spec.runtime_options.k8s_options.namespace,
            "inferenceservices",
            model_spec.model_details.name,
        )
        if self.inside_cluster is None:
            url = api_response["status"]["url"]
            o = urlparse(url)
            return o.hostname
        else:
            url = api_response["status"]["address"]["url"]
            o = urlparse(url)
            return o.hostname

    def get_url(self, model_spec: ModelSpec):
        if self.inside_cluster is None:
            ingress = create_ingress(model_spec)
            ingress_host_url = ingress.get_external_host_url(model_spec)
            return f"{ingress_host_url}" + model_spec.protocol.get_predict_path(model_spec.model_details)
        else:
            # TODO check why needed this here
            config.load_incluster_config()
            api_instance = client.CustomObjectsApi()
            api_response = api_instance.get_namespaced_custom_object_status(
                "serving.kubeflow.org",
                "v1beta1",
                model_spec.runtime_options.k8s_options.namespace,
                "inferenceservices",
                model_spec.model_details.name,
            )
            url = api_response["status"]["address"]["url"]
            # Hack until https://github.com/kubeflow/kfserving/issues/1483 fixed
            if (
                "serviceAccountName" in api_response["spec"]["predictor"]
                and api_response["spec"]["predictor"]["serviceAccountName"] == DefaultServiceAccountName
            ):
                url = url.replace("v1/models", "v2/models").replace(":predict", "/infer")
            return url



from typing import List, Tuple

import numpy as np
import pytest

from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.serve.metadata import ModelFramework
from tempo.serve.model import Model
from tempo.serve.utils import model, predictmethod


#
# Test single input output model with type annotations
#
@pytest.mark.parametrize(
    "v2_input, expected",
    [
        (
            {
                "inputs": [
                    {
                        "name": "a",
                        "datatype": "FP64",
                        "shape": [1, 4],
                        "data": [1, 2, 3, 4],
                    }
                ]
            },
            {
                "model_name": "custom",
                "outputs": [
                    {
                        "name": "output0",
                        "datatype": "FP64",
                        "data": [1.0, 2.0, 3.0, 4.0],
                        "shape": [1, 4],
                    }
                ],
            },
        )
    ],
)
def test_custom_model(v2_input, expected):
    @model(
        name="custom",
        protocol=KFServingV2Protocol(),
        platform=ModelFramework.Custom,
    )
    def custom_model(a: np.ndarray) -> np.ndarray:
        return a

    response = custom_model.request(v2_input)
    assert response == expected


#
# Test lambda function
#
@pytest.mark.parametrize("input, expected", [(np.array([[0, 0, 0, 1]]), np.array([[0, 0, 1]]))])
def test_lambda(input, expected):
    model = Model(
        name="test-iris-sklearn",
        platform=ModelFramework.Custom,
        model_func=lambda x: np.array([[0, 0, 1]]),
    )

    response = model(input)
    np.testing.assert_allclose(response, expected, atol=1e-2)


#
# Test single input output model with type annotations
#


@pytest.mark.parametrize(
    "v2_input, expected",
    [
        (
            {
                "inputs": [
                    {
                        "name": "a",
                        "datatype": "FP64",
                        "shape": [1, 4],
                        "data": [1, 2, 3, 4],
                    }
                ]
            },
            {
                "model_name": "custom",
                "outputs": [
                    {
                        "name": "output0",
                        "datatype": "FP64",
                        "data": [1.0, 2.0, 3.0, 4.0],
                        "shape": [1, 4],
                    }
                ],
            },
        )
    ],
)
def test_custom_model_decorator_types(v2_input, expected):
    @model(
        name="custom",
        platform=ModelFramework.Custom,
        inputs=np.ndarray,
        outputs=np.ndarray,
    )
    def custom_model_decorator_types(a):
        return a

    response = custom_model_decorator_types.request(v2_input)
    assert response == expected


#
# Test multi-headed input output model with type annotations with tuple return
#
@pytest.mark.parametrize(
    "v2_input, expected",
    [
        (
            {
                "inputs": [
                    {
                        "name": "a",
                        "datatype": "FP64",
                        "shape": [1, 4],
                        "data": [1, 2, 3, 4],
                    },
                    {"name": "b", "datatype": "FP64", "shape": [1, 2], "data": [4, 5]},
                ]
            },
            {
                "model_name": "multi-headed",
                "outputs": [
                    {
                        "name": "output0",
                        "datatype": "FP64",
                        "data": [1.0, 2.0, 3.0, 4.0],
                        "shape": [1, 4],
                    },
                    {
                        "name": "output1",
                        "datatype": "FP64",
                        "data": [4.0, 5.0],
                        "shape": [1, 2],
                    },
                ],
            },
        )
    ],
)
def test_custom_multiheaded_model_tuple(v2_input, expected):
    @model(
        name="multi-headed",
        platform=ModelFramework.Custom,
    )
    def custom_multiheaded_model_tuple(a: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return a, b

    response = custom_multiheaded_model_tuple.request(v2_input)
    assert response == expected


#
# Test multi-headed input output model with type annotations with list return
#
@pytest.mark.parametrize(
    "v2_input, expected",
    [
        (
            {
                "inputs": [
                    {
                        "name": "a",
                        "datatype": "FP64",
                        "shape": [1, 4],
                        "data": [1, 2, 3, 4],
                    },
                    {"name": "b", "datatype": "FP64", "shape": [1, 2], "data": [4, 5]},
                ]
            },
            {
                "model_name": "multi-headed",
                "outputs": [
                    {
                        "name": "output0",
                        "datatype": "FP64",
                        "data": [1.0, 2.0, 3.0, 4.0],
                        "shape": [1, 4],
                    },
                    {
                        "name": "output1",
                        "datatype": "FP64",
                        "data": [4.0, 5.0],
                        "shape": [1, 2],
                    },
                ],
            },
        )
    ],
)
def test_custom_multiheaded_model_list(v2_input, expected):
    @model(
        name="multi-headed",
        platform=ModelFramework.Custom,
    )
    def custom_multiheaded_model_list(a: np.ndarray, b: np.ndarray) -> List[np.ndarray]:
        return [a, b]

    response = custom_multiheaded_model_list.request(v2_input)
    assert response == expected


def test_class_func_class():
    @model(
        name="classifier",
        platform=ModelFramework.TempoPipeline,
    )
    class MyCustomModel:
        @predictmethod
        def predict(self, X: str) -> str:
            return X

    x = MyCustomModel()

    r = x.predict("hello")
    assert r == "hello"
    r = x("hello")
    assert r == "hello"



import abc
from pydoc import locate
from typing import Any

import attr

from tempo.serve.runtime import ModelSpec


@attr.s(auto_attribs=True)
class Ingress(abc.ABC):
    @abc.abstractmethod
    def get_external_host_url(self, model_spec: ModelSpec) -> str:
        pass


def create_ingress(model_spec: ModelSpec) -> Ingress:
    cls: Any = locate(model_spec.runtime_options.ingress_options.ingress)
    return cls()



from typing import Any, Callable

from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.serve.base import BaseModel
from tempo.serve.constants import ModelDataType
from tempo.serve.metadata import ModelFramework, RuntimeOptions
from tempo.serve.protocol import Protocol


class Model(BaseModel):
    def __init__(
        self,
        name: str,
        protocol: Protocol = KFServingV2Protocol(),
        local_folder: str = None,
        uri: str = None,
        platform: ModelFramework = None,
        inputs: ModelDataType = None,
        outputs: ModelDataType = None,
        model_func: Callable[..., Any] = None,
        conda_env: str = None,
        runtime_options: RuntimeOptions = RuntimeOptions(),
    ):
        super().__init__(
            name,
            # TODO: Should we unify names?
            user_func=model_func,
            local_folder=local_folder,
            uri=uri,
            platform=platform,
            inputs=inputs,
            outputs=outputs,
            conda_env=conda_env,
            protocol=protocol,
            runtime_options=runtime_options,
        )



import json

from tempo.kfserving.protocol import KFServingV1Protocol, KFServingV2Protocol
from tempo.seldon.constants import MLSERVER_IMAGE
from tempo.serve.constants import ENV_TEMPO_RUNTIME_OPTIONS
from tempo.serve.metadata import ModelDetails, ModelFramework, RuntimeOptions
from tempo.serve.runtime import ModelSpec

DefaultHTTPPort = "9000"
DefaultGRPCPort = "9500"

DefaultModelsPath = "/mnt/models"
DefaultServiceAccountName = "tempo-pipeline"


def get_container_spec(model_details: ModelSpec) -> dict:
    if model_details.model_details.platform == ModelFramework.TempoPipeline:
        return _V2ContainerFactory.get_container_spec(model_details.model_details, model_details.runtime_options)

    if isinstance(model_details.protocol, KFServingV2Protocol):
        return _V2ContainerFactory.get_container_spec(model_details.model_details, model_details.runtime_options)

    return _V1ContainerFactory.get_container_spec(model_details.model_details)


class _V1ContainerFactory:
    Images = {
        ModelFramework.SKLearn: "seldonio/sklearnserver:1.6.0-dev",
        ModelFramework.XGBoost: "seldonio/xgboostserver:1.6.0-dev",
        ModelFramework.Tensorflow: "tensorflow/serving:2.1.0",
    }

    @classmethod
    def get_container_spec(cls, model_details: ModelDetails) -> dict:
        model_image = cls.Images[model_details.platform]

        if model_details.platform == ModelFramework.Tensorflow:
            return {
                "image": model_image,
                "command": [
                    "--rest_api_port=" + DefaultHTTPPort,
                    "--model_name=" + model_details.name,
                    "--model_base_path=" + DefaultModelsPath,
                ],
            }
        else:
            parameters = [{"name": "model_uri", "value": DefaultModelsPath, "type": "STRING"}]
            env = {"PREDICTIVE_UNIT_PARAMETERS": json.dumps(parameters)}

            return {
                "image": model_image,
                "environment": env,
            }


class _V2ContainerFactory:
    MLServerImage = MLSERVER_IMAGE

    MLServerRuntimes = {
        ModelFramework.SKLearn: "mlserver_sklearn.SKLearnModel",
        ModelFramework.XGBoost: "mlserver_xgboost.XGBoostModel",
        ModelFramework.Custom: "tempo.mlserver.InferenceRuntime",
        ModelFramework.TempoPipeline: "tempo.mlserver.InferenceRuntime",
    }

    @classmethod
    def get_container_spec(cls, model_details: ModelDetails, runtime_options: RuntimeOptions) -> dict:
        mlserver_runtime = cls.MLServerRuntimes[model_details.platform]

        env = {
            "MLSERVER_HTTP_PORT": DefaultHTTPPort,
            "MLSERVER_GRPC_PORT": DefaultGRPCPort,
            "MLSERVER_MODEL_IMPLEMENTATION": mlserver_runtime,
            "MLSERVER_MODEL_NAME": model_details.name,
            "MLSERVER_MODEL_URI": DefaultModelsPath,
            ENV_TEMPO_RUNTIME_OPTIONS: json.dumps(runtime_options.dict()),
        }

        return {
            "image": cls.MLServerImage,
            "environment": env,
        }


class KubernetesSpec:
    Implementations = {
        ModelFramework.SKLearn: "SKLEARN_SERVER",
        ModelFramework.XGBoost: "XGBOOST_SERVER",
        ModelFramework.MLFlow: "MLFLOW_SERVER",
        ModelFramework.Tensorflow: "TENSORFLOW_SERVER",
        ModelFramework.PyTorch: "TRITON_SERVER",
        ModelFramework.ONNX: "TRITON_SERVER",
        ModelFramework.TensorRT: "TRITON_SERVER",
        # TODO: We need to set an implementation in order to get the init
        # container injected into the spec
        ModelFramework.TempoPipeline: "TRITON_SERVER",
    }

    def __init__(
        self,
        model_details: ModelSpec,
    ):
        self._details = model_details

    @property
    def spec(self) -> dict:
        predictor = self._get_predictor()
        protocol = self._get_spec_protocol()

        return {
            "apiVersion": "machinelearning.seldon.io/v1",
            "kind": "SeldonDeployment",
            "metadata": {
                "name": self._details.model_details.name,
                "namespace": self._details.runtime_options.k8s_options.namespace,
            },
            "spec": {"protocol": protocol, "predictors": [predictor]},
        }

    def _get_predictor(self) -> dict:
        # TODO: We need to insert `type: MODEL`, otherwise the validation
        # webhook complains
        graph = {
            "modelUri": self._details.model_details.uri,
            "name": self._details.model_details.name,
            "type": "MODEL",
        }

        if self._details.runtime_options.k8s_options.authSecretName:
            graph["envSecretRefName"] = self._details.runtime_options.k8s_options.authSecretName

        if self._details.model_details.platform in self.Implementations:
            model_implementation = self.Implementations[self._details.model_details.platform]
            graph["implementation"] = model_implementation

        if self._details.model_details.platform == ModelFramework.TempoPipeline:
            serviceAccountName = self._details.runtime_options.k8s_options.serviceAccountName
            if serviceAccountName is None:
                serviceAccountName = DefaultServiceAccountName
            graph["serviceAccountName"] = serviceAccountName

        predictor = {
            "graph": graph,
            "name": "default",
            "replicas": self._details.runtime_options.k8s_options.replicas,
        }

        if self._details.model_details.platform == ModelFramework.TempoPipeline:
            predictor["componentSpecs"] = self._get_component_specs()

        return predictor

    def _get_component_specs(self) -> list:
        container_spec = get_container_spec(self._details)
        container_env = [{"name": name, "value": value} for name, value in container_spec["environment"].items()]

        return [
            {
                "spec": {
                    "containers": [
                        {
                            "name": self._details.model_details.name,
                            "image": container_spec["image"],
                            "env": container_env,
                            # TODO: Necessary to override Triton defaults (see
                            # note above)
                            "args": [],
                        }
                    ]
                }
            }
        ]

    def _get_spec_protocol(self) -> str:
        if isinstance(self._details.protocol, KFServingV2Protocol):
            return "kfserving"

        if isinstance(self._details.protocol, KFServingV1Protocol):
            return "tensorflow"

        return "seldon"



import json
import os
import time
from typing import Any, Dict, Optional

import requests
import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException

from tempo.kfserving.endpoint import Endpoint
from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.seldon.constants import MLSERVER_IMAGE
from tempo.seldon.specs import DefaultModelsPath, DefaultServiceAccountName
from tempo.serve.constants import ENV_TEMPO_RUNTIME_OPTIONS
from tempo.serve.metadata import ModelFramework, RuntimeOptions
from tempo.serve.remote import Remote
from tempo.serve.runtime import ModelSpec, Runtime
from tempo.utils import logger

DefaultHTTPPort = "8080"
DefaultGRPCPort = "9000"

ENV_K8S_SERVICE_HOST = "KUBERNETES_SERVICE_HOST"

Implementations = {
    ModelFramework.SKLearn: "sklearn",
    ModelFramework.XGBoost: "xgboost",
    ModelFramework.Tensorflow: "tensorflow",
    ModelFramework.PyTorch: "triton",
    ModelFramework.ONNX: "triton",
    ModelFramework.TensorRT: "triton",
}


class KFServingKubernetesRuntime(Runtime, Remote):
    def __init__(self, runtime_options: Optional[RuntimeOptions] = None):
        if runtime_options:
            runtime_options.runtime = "tempo.kfserving.KFServingKubernetesRuntime"
        super().__init__(runtime_options)

    def _inside_cluster(self):
        return os.getenv(ENV_K8S_SERVICE_HOST)

    def create_k8s_client(self):
        if self._inside_cluster():
            logger.debug("Loading cluster local config")
            config.load_incluster_config()
            return True
        else:
            logger.debug("Loading external kubernetes config")
            config.load_kube_config()
            return False

    def get_endpoint_spec(self, model_spec: ModelSpec) -> str:
        endpoint = Endpoint()
        return endpoint.get_url(model_spec)

    def get_headers(self, model_spec: ModelSpec) -> Dict[str, str]:
        if not self._inside_cluster():
            endpoint = Endpoint()
            service_host = endpoint.get_service_host(model_spec)
            return {"Host": service_host}
        else:
            return {}

    def remote(self, model_spec: ModelSpec, *args, **kwargs) -> Any:
        req = model_spec.protocol.to_protocol_request(*args, **kwargs)
        endpoint = self.get_endpoint_spec(model_spec)
        logger.debug("Endpoint is", endpoint)
        headers = self.get_headers(model_spec)
        logger.debug("Headers are", headers)
        response_raw = requests.post(
            endpoint, json=req, headers=headers, verify=model_spec.runtime_options.ingress_options.verify_ssl
        )
        if response_raw.status_code == 200:
            return model_spec.protocol.from_protocol_response(response_raw.json(), model_spec.model_details.outputs)
        else:
            raise ValueError("Bad return code", response_raw.status_code, response_raw.text)

    def undeploy_spec(self, model_spec: ModelSpec):
        self.create_k8s_client()
        api_instance = client.CustomObjectsApi()
        api_instance.delete_namespaced_custom_object(
            "serving.kubeflow.org",
            "v1beta1",
            model_spec.runtime_options.k8s_options.namespace,
            "inferenceservices",
            model_spec.model_details.name,
            body=client.V1DeleteOptions(propagation_policy="Foreground"),
        )

    def deploy_spec(self, model_spec: ModelSpec):
        spec = self._get_spec(model_spec)
        logger.debug(model_spec)
        self.create_k8s_client()
        api_instance = client.CustomObjectsApi()

        try:
            existing = api_instance.get_namespaced_custom_object(
                "serving.kubeflow.org",
                "v1beta1",
                model_spec.runtime_options.k8s_options.namespace,
                "inferenceservices",
                model_spec.model_details.name,
            )
            spec["metadata"]["resourceVersion"] = existing["metadata"]["resourceVersion"]
            api_instance.replace_namespaced_custom_object(
                "serving.kubeflow.org",
                "v1beta1",
                model_spec.runtime_options.k8s_options.namespace,
                "inferenceservices",
                model_spec.model_details.name,
                spec,
            )
        except ApiException as e:
            if e.status == 404:
                api_instance.create_namespaced_custom_object(
                    "serving.kubeflow.org",
                    "v1beta1",
                    model_spec.runtime_options.k8s_options.namespace,
                    "inferenceservices",
                    spec,
                )
            else:
                raise e

    def wait_ready_spec(self, model_spec: ModelSpec, timeout_secs=None) -> bool:
        self.create_k8s_client()
        ready = False
        t0 = time.time()
        while not ready:
            api_instance = client.CustomObjectsApi()
            existing = api_instance.get_namespaced_custom_object(
                "serving.kubeflow.org",
                "v1beta1",
                model_spec.runtime_options.k8s_options.namespace,
                "inferenceservices",
                model_spec.model_details.name,
            )
            default_ready = False
            routes_ready = False
            conf_ready = False
            ingress_ready = False
            if "status" in existing and "conditions" in existing["status"]:
                for item in existing["status"]["conditions"]:
                    if item["type"] == "PredictorReady":
                        default_ready = item["status"] == "True"
                    elif item["type"] == "PredictorRouteReady":
                        routes_ready = item["status"] == "True"
                    elif item["type"] == "IngressReady":
                        ingress_ready = item["status"] == "True"
                    elif item["type"] == "PredictorConfigurationReady":
                        conf_ready = item["status"] == "True"
                ready = default_ready and routes_ready and ingress_ready and conf_ready
            if timeout_secs is not None:
                t1 = time.time()
                if t1 - t0 > timeout_secs:
                    return ready
            time.sleep(1)
        return ready

    def _get_spec(self, model_spec: ModelSpec) -> dict:
        if model_spec.model_details.platform == ModelFramework.TempoPipeline:
            serviceAccountName = model_spec.runtime_options.k8s_options.serviceAccountName
            if serviceAccountName is None:
                serviceAccountName = DefaultServiceAccountName
            return {
                "apiVersion": "serving.kubeflow.org/v1beta1",
                "kind": "InferenceService",
                "metadata": {
                    "name": model_spec.model_details.name,
                    "namespace": model_spec.runtime_options.k8s_options.namespace,
                },
                "spec": {
                    "predictor": {
                        "serviceAccountName": serviceAccountName,
                        "containers": [
                            {
                                "image": MLSERVER_IMAGE,
                                "name": "mlserver",
                                "env": [
                                    {
                                        "name": "STORAGE_URI",
                                        "value": model_spec.model_details.uri,
                                    },
                                    {
                                        "name": "MLSERVER_HTTP_PORT",
                                        "value": DefaultHTTPPort,
                                    },
                                    {
                                        "name": "MLSERVER_GRPC_PORT",
                                        "value": DefaultGRPCPort,
                                    },
                                    {
                                        "name": "MLSERVER_MODEL_IMPLEMENTATION",
                                        "value": "tempo.mlserver.InferenceRuntime",
                                    },
                                    {
                                        "name": "MLSERVER_MODEL_NAME",
                                        "value": model_spec.model_details.name,
                                    },
                                    {
                                        "name": "MLSERVER_MODEL_URI",
                                        "value": DefaultModelsPath,
                                    },
                                    {
                                        "name": ENV_TEMPO_RUNTIME_OPTIONS,
                                        "value": json.dumps(model_spec.runtime_options.dict()),
                                    },
                                ],
                            },
                        ],
                    },
                },
            }
        elif model_spec.model_details.platform in Implementations:
            model_implementation = Implementations[model_spec.model_details.platform]
            spec: Dict = {
                "apiVersion": "serving.kubeflow.org/v1beta1",
                "kind": "InferenceService",
                "metadata": {
                    "name": model_spec.model_details.name,
                    "namespace": model_spec.runtime_options.k8s_options.namespace,
                },
                "spec": {
                    "predictor": {
                        model_implementation: {"storageUri": model_spec.model_details.uri},
                    },
                },
            }
            if model_spec.runtime_options.k8s_options.serviceAccountName is not None:
                spec["spec"]["predictor"][
                    "serviceAccountName"
                ] = model_spec.runtime_options.k8s_options.serviceAccountName
            if isinstance(model_spec.protocol, KFServingV2Protocol):
                spec["spec"]["predictor"][model_implementation]["protocolVersion"] = "v2"
            return spec
        else:
            raise ValueError("Can't create spec for implementation ", model_spec.model_details.platform)

    def to_k8s_yaml_spec(self, model_spec: ModelSpec) -> str:
        d = self._get_spec(model_spec)
        return yaml.safe_dump(d)



import numpy as np

from tempo.serve.metadata import ModelFramework
from tempo.serve.model import Model


def test_default_types():
    model = Model("mymodel", local_folder="", uri="", platform=ModelFramework.SKLearn)
    assert len(model.model_spec.model_details.inputs) == 1
    assert model.model_spec.model_details.inputs[0] == np.ndarray
    assert len(model.model_spec.model_details.outputs) == 1
    assert model.model_spec.model_details.outputs[0] == np.ndarray



import os
import re
import uuid
from subprocess import run
from tempfile import NamedTemporaryFile
from typing import Any, Optional

import cloudpickle
import conda_pack
import rclone
import yaml

from tempo.conf import settings
from tempo.serve.constants import DefaultModelFilename, DefaultRemoteFilename, MLServerEnvDeps
from tempo.utils import logger


def save(tempo_artifact: Any, save_env=True):
    model = tempo_artifact.get_tempo()
    model.set_remote(True)
    model.save(save_env=save_env)
    model.set_remote(False)


def save_custom(pipeline, file_path: str) -> str:
    with open(file_path, "wb") as file:
        cloudpickle.dump(pipeline, file)

    return file_path


def load(folder: str):
    file_path_pkl = os.path.join(folder, DefaultModelFilename)
    return load_custom(file_path_pkl)


def load_custom(file_path: str):
    with open(file_path, "rb") as file:
        return cloudpickle.load(file)


def load_remote(folder: str):
    file_path = os.path.join(folder, DefaultRemoteFilename)
    with open(file_path, "rb") as file:
        return cloudpickle.load(file)


def _get_env(conda_env_file_path: str = None, env_name: str = None) -> dict:
    if conda_env_file_path:
        with open(conda_env_file_path) as file:
            env = yaml.safe_load(file)
            if not _has_required_deps(env):
                raise ValueError(f"conda.yaml does not contain {MLServerEnvDeps}")
            else:
                logger.info("Using found conda.yaml")
    else:
        env = _get_environment(env_name=env_name)
        env = _add_required_deps(env)
    return env


def save_environment(conda_pack_file_path: str, conda_env_file_path: str = None, env_name: str = None):
    if env_name:
        _pack_environment(env_name, conda_pack_file_path)
    else:
        env = _get_env(conda_env_file_path, env_name)
        _create_and_pack_environment(env=env, file_path=conda_pack_file_path)


def _get_environment(env_name: str = None) -> dict:
    cmd = "conda env export"

    if env_name:
        cmd += f" --name {env_name}"

    proc = run(cmd, shell=True, check=True, capture_output=True)
    return yaml.safe_load(proc.stdout)


def _has_required_deps(env: dict) -> bool:
    if "dependencies" not in env:
        return False

    dependencies = env["dependencies"]
    pip_deps = _get_pip_deps(dependencies)
    if not pip_deps:
        return False

    for dep in MLServerEnvDeps:
        parts = re.split(r"==|>=|<=|~=|!=|>|<|==:", dep)
        module = parts[0]
        r = re.compile(fr"{module}$|({module}((==|>=|<=|~=|!=|>|<|==:)[0-9]+\.[0-9]+.[0-9]+))")
        newlist = list(filter(r.match, pip_deps["pip"]))
        if len(newlist) == 0:
            return False

    return True


def _add_required_deps(env: dict) -> dict:
    if "dependencies" not in env:
        env["dependencies"] = []

    dependencies = env["dependencies"]
    pip_deps = _get_pip_deps(dependencies)
    if not pip_deps:
        pip_deps = {"pip": []}
        dependencies.append(pip_deps)

    for dep in MLServerEnvDeps:
        parts = re.split(r"==|>=|<=|~=|!=|>|<|==:", dep)
        module = parts[0]
        r = re.compile(fr"{module}$|({module}((==|>=|<=|~=|!=|>|<|==:)[0-9]+\.[0-9]+.[0-9]+))")
        newlist = list(filter(r.match, pip_deps["pip"]))
        if len(newlist) == 0:
            pip_deps["pip"].extend(MLServerEnvDeps)

    return env


def _get_pip_deps(dependencies: dict) -> Optional[dict]:
    for dep in dependencies:
        if isinstance(dep, dict) and "pip" in dep:
            # If entry is a dict, and has a `pip` key that's the one
            return dep

    return None


def _pack_environment(env_name: str, file_path: str):
    logger.info("packing conda environment from %s", env_name)
    # Pack environment
    conda_pack.pack(
        name=env_name,
        output=file_path,
        force=True,
        verbose=True,
        ignore_editable_packages=False,
        ignore_missing_files=True,
    )


def _create_and_pack_environment(env: dict, file_path: str):
    with NamedTemporaryFile(mode="w", suffix=".yml") as file:
        # TODO: Save copy of environment.yaml alongside tarball
        yaml.safe_dump(env, file)

        # Create env
        tmp_env_path = file.name
        tmp_env_name = f"tempo-{uuid.uuid4()}"
        cmd = f"conda env create --name {tmp_env_name} --file {tmp_env_path}"
        logger.info("Creating conda env with: %s", cmd)
        run(cmd, shell=True, check=True)

        try:
            _pack_environment(tmp_env_name, file_path)
        finally:
            # Remove environment
            cmd = f"conda remove --name {tmp_env_name} --all --yes"
            logger.info("Removing conda env with: %s", cmd)
            run(cmd, shell=True, check=True)


def _load_rclone_cfg() -> str:
    with open(settings.rclone_cfg, "r") as f:
        return f.read()


def upload(tempo_artifact: Any):
    "Upload local to remote using rclone"
    model = tempo_artifact.get_tempo()
    local_path = model.details.local_folder
    remote_uri = model.details.uri
    logger.info("Uploading %s to %s", local_path, remote_uri)
    rclone.with_config(_load_rclone_cfg()).copy(local_path, remote_uri, flags=["-P"])


def download(tempo_artifact: Any):
    "Download remote to local using rclone"
    model = tempo_artifact.get_tempo()
    local_path = model.details.local_folder
    remote_uri = model.details.uri
    logger.info("Downloading %s to %s", remote_uri, local_path)
    rclone.with_config(_load_rclone_cfg()).copy(remote_uri, local_path, flags=["-P"])



import os
import time
import uuid
from typing import Generator

import pytest
from kubernetes import client, config
from kubernetes.utils.create_from_yaml import create_from_yaml

from tempo import Model, Pipeline
from tempo.kfserving import KFServingV2Protocol
from tempo.seldon import SeldonKubernetesRuntime
from tempo.serve.metadata import KubernetesOptions

from ...conftest import TESTDATA_PATH

K8S_NAMESPACE_PREFIX = "test-tempo-"


@pytest.fixture
def namespace() -> Generator[str, None, None]:
    unique_id = str(uuid.uuid4().fields[-1])[:5]
    ns_name = f"{K8S_NAMESPACE_PREFIX}{unique_id}"

    config.load_kube_config()
    core_api = client.CoreV1Api()

    namespace = client.V1Namespace(metadata=client.V1ObjectMeta(name=ns_name))
    core_api.create_namespace(namespace)

    # Create ServiceAccount and RBAC
    rbac_path = os.path.join(TESTDATA_PATH, "tempo-pipeline-rbac.yaml")
    api_client = client.ApiClient()
    create_from_yaml(api_client, rbac_path, namespace=ns_name)

    yield ns_name

    core_api.delete_namespace(ns_name)


@pytest.fixture
def runtime(namespace: str) -> SeldonKubernetesRuntime:
    return SeldonKubernetesRuntime(k8s_options=KubernetesOptions(namespace=namespace))


@pytest.fixture
def runtime_v2(namespace: str) -> SeldonKubernetesRuntime:
    return SeldonKubernetesRuntime(
        k8s_options=KubernetesOptions(namespace=namespace),
        protocol=KFServingV2Protocol(),
    )


@pytest.fixture
def sklearn_model(sklearn_model: Model, runtime: SeldonKubernetesRuntime) -> Generator[Model, None, None]:
    sklearn_model.set_runtime(runtime)

    sklearn_model.deploy()
    sklearn_model.wait_ready(timeout_secs=60)

    yield sklearn_model

    sklearn_model.undeploy()


@pytest.fixture
def xgboost_model(xgboost_model: Model, runtime: SeldonKubernetesRuntime) -> Generator[Model, None, None]:
    xgboost_model.set_runtime(runtime)
    xgboost_model.deploy()
    xgboost_model.wait_ready(timeout_secs=60)

    yield xgboost_model

    xgboost_model.undeploy()


@pytest.fixture
def inference_pipeline(
    inference_pipeline: Pipeline,
    runtime: SeldonKubernetesRuntime,
    runtime_v2: SeldonKubernetesRuntime,
) -> Generator[Pipeline, None, None]:
    inference_pipeline.set_runtime(runtime_v2)

    for model in inference_pipeline._models:
        model.set_runtime(runtime)

    inference_pipeline.save(save_env=False)
    inference_pipeline.upload()
    inference_pipeline.deploy()
    # TODO: Fix wait_ready for pipelines
    time.sleep(60)
    inference_pipeline.wait_ready(timeout_secs=60)

    yield inference_pipeline

    inference_pipeline.undeploy()



import os
import time
from typing import Any, Optional

import requests
import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException

from tempo.seldon.endpoint import Endpoint
from tempo.seldon.specs import KubernetesSpec
from tempo.serve.constants import ENV_K8S_SERVICE_HOST
from tempo.serve.metadata import RuntimeOptions
from tempo.serve.remote import Remote
from tempo.serve.runtime import ModelSpec, Runtime
from tempo.utils import logger


class SeldonKubernetesRuntime(Runtime, Remote):
    def __init__(self, runtime_options: Optional[RuntimeOptions] = None):
        if runtime_options:
            runtime_options.runtime = "tempo.seldon.SeldonKubernetesRuntime"
        super().__init__(runtime_options)

    def create_k8s_client(self):
        inside_cluster = os.getenv(ENV_K8S_SERVICE_HOST)
        if inside_cluster:
            logger.debug("Loading cluster local config")
            config.load_incluster_config()
        else:
            logger.debug("Loading external kubernetes config")
            config.load_kube_config()

    def get_endpoint_spec(self, model_spec: ModelSpec) -> str:
        self.create_k8s_client()
        endpoint = Endpoint()
        return endpoint.get_url(model_spec)

    def remote(self, model_spec: ModelSpec, *args, **kwargs) -> Any:
        req = model_spec.protocol.to_protocol_request(*args, **kwargs)
        endpoint = self.get_endpoint_spec(model_spec)
        logger.debug("Endpoint is ", endpoint)
        response_raw = requests.post(endpoint, json=req, verify=model_spec.runtime_options.ingress_options.verify_ssl)
        return model_spec.protocol.from_protocol_response(response_raw.json(), model_spec.model_details.outputs)

    def undeploy_spec(self, model_spec: ModelSpec):
        self.create_k8s_client()
        api_instance = client.CustomObjectsApi()
        api_instance.delete_namespaced_custom_object(
            "machinelearning.seldon.io",
            "v1",
            model_spec.runtime_options.k8s_options.namespace,
            "seldondeployments",
            model_spec.model_details.name,
            body=client.V1DeleteOptions(propagation_policy="Foreground"),
        )

    def deploy_spec(self, model_spec: ModelSpec):
        self.create_k8s_client()
        k8s_specer = KubernetesSpec(model_spec)
        k8s_spec = k8s_specer.spec
        logger.debug(k8s_spec)

        api_instance = client.CustomObjectsApi()

        try:
            existing = api_instance.get_namespaced_custom_object(
                "machinelearning.seldon.io",
                "v1",
                model_spec.runtime_options.k8s_options.namespace,
                "seldondeployments",
                model_spec.model_details.name,
            )
            k8s_spec["metadata"]["resourceVersion"] = existing["metadata"]["resourceVersion"]
            api_instance.replace_namespaced_custom_object(
                "machinelearning.seldon.io",
                "v1",
                model_spec.runtime_options.k8s_options.namespace,
                "seldondeployments",
                model_spec.model_details.name,
                k8s_spec,
            )
        except ApiException as e:
            if e.status == 404:
                api_instance.create_namespaced_custom_object(
                    "machinelearning.seldon.io",
                    "v1",
                    model_spec.runtime_options.k8s_options.namespace,
                    "seldondeployments",
                    k8s_spec,
                )
            else:
                raise e

    def wait_ready_spec(self, model_spec: ModelSpec, timeout_secs=None) -> bool:
        self.create_k8s_client()
        ready = False
        t0 = time.time()
        while not ready:
            api_instance = client.CustomObjectsApi()
            existing = api_instance.get_namespaced_custom_object(
                "machinelearning.seldon.io",
                "v1",
                model_spec.runtime_options.k8s_options.namespace,
                "seldondeployments",
                model_spec.model_details.name,
            )
            if "status" in existing and "state" in existing["status"]:
                ready = existing["status"]["state"] == "Available"
            if timeout_secs is not None:
                t1 = time.time()
                if t1 - t0 > timeout_secs:
                    return ready
        return ready

    def to_k8s_yaml_spec(self, model_spec: ModelSpec) -> str:
        k8s_spec = KubernetesSpec(model_spec)
        return yaml.safe_dump(k8s_spec.spec)



from tempo.seldon.docker import SeldonDockerRuntime
from tempo.serve.model import Model


def test_launch_tensorflow(cifar10_model: Model, runtime: SeldonDockerRuntime):
    container = runtime._get_container(cifar10_model.model_spec)
    assert container.status == "running"



from enum import Enum
from typing import List, Optional, Type, Union

from pydantic import BaseModel


class ModelFramework(Enum):
    SKLearn = "sklearn"
    XGBoost = "xgboost"
    MLFlow = "mlflow"
    Tensorflow = "tensorflow"
    PyTorch = "pytorch"
    ONNX = "ONNX"
    TensorRT = "tensorrt"
    Alibi = "alibi"
    Custom = "custom"
    TempoPipeline = "tempo"


class ModelDataArg(BaseModel):
    ty: Type
    name: Optional[str] = None


class ModelDataArgs(BaseModel):

    args: List[ModelDataArg]

    def _get_type_by_name(self, name: str) -> Optional[Type]:
        for arg in self.args:
            if name == arg.name:
                return arg.ty
        return None

    def __getitem__(self, idx: Union[str, int]) -> Optional[Type]:
        if isinstance(idx, str):
            return self._get_type_by_name(idx)

        if idx < len(self.args):
            # NOTE: `idx` here must be an int
            return self.args[idx].ty

        return None

    def __len__(self):
        return len(self.args)


class ModelDetails(BaseModel):
    name: str
    local_folder: str
    uri: str
    platform: ModelFramework
    inputs: ModelDataArgs
    outputs: ModelDataArgs


class KubernetesOptions(BaseModel):
    replicas: int = 1
    namespace = "default"
    minReplicas: Optional[int] = None
    maxReplicas: Optional[int] = None
    authSecretName: Optional[str] = None
    serviceAccountName: Optional[str] = None
    defaultRuntime: str = "tempo.seldon.SeldonKubernetesRuntime"


class DockerOptions(BaseModel):
    defaultRuntime: str = "tempo.seldon.SeldonDockerRuntime"


class IngressOptions(BaseModel):
    ingress: str = "tempo.ingress.istio.IstioIngress"
    ssl: bool = False
    verify_ssl: bool = True


class RuntimeOptions(BaseModel):
    runtime: Optional[str] = None
    docker_options: DockerOptions = DockerOptions()
    k8s_options: KubernetesOptions = KubernetesOptions()
    ingress_options: IngressOptions = IngressOptions()



import os
import socket
import time
from typing import Any, Optional, Tuple

import docker
import requests
from docker.client import DockerClient
from docker.errors import NotFound
from docker.models.containers import Container

from tempo.seldon.specs import DefaultHTTPPort, DefaultModelsPath, get_container_spec
from tempo.serve.metadata import RuntimeOptions
from tempo.serve.remote import Remote
from tempo.serve.runtime import ModelSpec, Runtime

DefaultNetworkName = "tempo"


class SeldonDockerRuntime(Runtime, Remote):
    def __init__(self, runtime_options: Optional[RuntimeOptions] = None):
        if runtime_options:
            runtime_options.runtime = "tempo.seldon.SeldonDockerRuntime"
        super().__init__(runtime_options)

    def _get_host_ip_port(self, model_details: ModelSpec) -> Tuple[str, str]:
        container = self._get_container(model_details)
        port_index = self._get_port_index()
        host_ports = container.ports[port_index]

        host_ip = host_ports[0]["HostIp"]
        host_port = host_ports[0]["HostPort"]
        return host_ip, host_port

    def get_endpoint_spec(self, model_spec: ModelSpec) -> str:
        predict_path = model_spec.protocol.get_predict_path(model_spec.model_details)

        if self._is_inside_docker():
            # If inside Docker, use internal networking
            return f"http://{model_spec.model_details.name}:{DefaultHTTPPort}{predict_path}"

        host_ip, host_port = self._get_host_ip_port(model_spec)

        return f"http://{host_ip}:{host_port}{predict_path}"

    def remote(self, model_spec: ModelSpec, *args, **kwargs) -> Any:
        req = model_spec.protocol.to_protocol_request(*args, **kwargs)
        endpoint = self.get_endpoint_spec(model_spec)
        response_raw = requests.post(endpoint, json=req)
        return model_spec.protocol.from_protocol_response(response_raw.json(), model_spec.model_details.outputs)

    def deploy_spec(self, model_details: ModelSpec):
        try:
            container = self._get_container(model_details)
            if container.status == "running":
                # If container already exists and is running, nothing to do
                # here
                return

            # Remove before re-deploying
            container.remove(force=True)
            self._run_container(model_details)
        except docker.errors.NotFound:
            self._run_container(model_details)

    def _run_container(self, model_details: ModelSpec):
        docker_client = docker.from_env()
        uid = os.getuid()

        container_index = self._get_port_index()
        model_folder = model_details.model_details.local_folder
        container_spec = get_container_spec(model_details)
        self._create_network(docker_client)

        docker_client.containers.run(
            name=self._get_container_name(model_details),
            ports={container_index: self._get_available_port()},
            volumes={model_folder: {"bind": DefaultModelsPath, "mode": "ro"}},
            detach=True,
            network=DefaultNetworkName,
            user=uid,
            **container_spec,
        )

    def _create_network(self, docker_client: DockerClient, network_name=DefaultNetworkName):
        try:
            docker_client.networks.get(network_id=network_name)
        except NotFound:
            docker_client.networks.create(name=DefaultNetworkName)

    def _get_port_index(self):
        return f"{DefaultHTTPPort}/tcp"

    def wait_ready_spec(self, model_spec: ModelSpec, timeout_secs=None) -> bool:
        host_ip, host_port = self._get_host_ip_port(model_spec)
        ready = False
        t0 = time.time()
        while not ready:
            container = self._get_container(model_spec)
            if container.status == "running":
                # TODO: Use status_path to wait until container is ready
                #  status_path = self.protocol.get_status_path(model_details)
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    s.connect((host_ip, int(host_port)))
                    return True
                except ConnectionRefusedError:
                    pass
                finally:
                    s.close()
                # url = f"http://{host_ip}:{host_port}{status_path}"
                # print(url)
                # r = requests.get(url)
                # if r.status_code == 200:
                #    return True
            if timeout_secs is not None:
                t1 = time.time()
                if t1 - t0 > timeout_secs:
                    return ready

        return False

    def _get_available_port(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
        return port

    def undeploy_spec(self, model_spec: ModelSpec):
        container = self._get_container(model_spec)
        container.remove(force=True)

    def _get_container(self, model_details: ModelSpec) -> Container:
        container_name = self._get_container_name(model_details)

        docker_client = docker.from_env()
        container = docker_client.containers.get(container_name)

        return container

    def _get_container_name(self, model_details: ModelSpec):
        return model_details.model_details.name

    def _is_inside_docker(self) -> bool:
        # From https://stackoverflow.com/a/48710609/5015573
        path = "/proc/self/cgroup"
        return os.path.exists("/.dockerenv") or os.path.isfile(path) and any("docker" in line for line in open(path))

    def to_k8s_yaml_spec(self, model_spec: ModelSpec) -> str:
        raise NotImplementedError()



import pytest

from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.serve.metadata import ModelDataArg, ModelDataArgs


@pytest.mark.parametrize(
    "data, expected",
    [("abc", [97, 98, 99]), ({"a": 1}, [123, 39, 97, 39, 58, 32, 49, 125])],
)
def test_v2_from_any(data, expected):
    d = KFServingV2Protocol.create_v2_from_any(data, "a")
    assert d["name"] == "a"
    assert d["data"] == expected
    assert d["datatype"] == "BYTES"


@pytest.mark.parametrize(
    "data, ty, expected",
    [([97, 98, 99], str, "abc"), ([123, 39, 97, 39, 58, 32, 49, 125], dict, {"a": 1})],
)
def test_convert_from_bytes(data, ty, expected):
    output = {"data": data}
    res = KFServingV2Protocol.convert_from_bytes(output, ty)
    assert res == expected


def test_v2_from_protocol_response():
    res = {"outputs": [{"name": "a", "data": [97, 98, 99], "datatype": "BYTES"}]}
    modelTyArgs = ModelDataArgs(args=[ModelDataArg(ty=str, name=None)])
    v2 = KFServingV2Protocol()
    res = v2.from_protocol_response(res, modelTyArgs)
    print(res)



import os

from kubernetes import client, config

from tempo.serve.ingress import create_ingress
from tempo.serve.runtime import ModelSpec
from tempo.utils import logger

ENV_K8S_SERVICE_HOST = "KUBERNETES_SERVICE_HOST"
ISTIO_GATEWAY = "istio"


class Endpoint(object):
    """A Model Endpoint

    Only handles istio and seldon at present.

    """

    def __init__(self):
        self.inside_cluster = os.getenv(ENV_K8S_SERVICE_HOST)
        try:
            if self.inside_cluster:
                logger.debug("Loading cluster local config")
                config.load_incluster_config()
            else:
                logger.debug("Loading external kubernetes config")
                config.load_kube_config()
        except Exception:
            logger.warning("Failed to load kubeconfig. Only local mode is possible.")

    def get_url(self, model_spec: ModelSpec):
        if self.inside_cluster is None:
            ingress = create_ingress(model_spec)
            ingress_host_url = ingress.get_external_host_url(model_spec)
            return (
                f"{ingress_host_url}"
                + f"/seldon/{model_spec.runtime_options.k8s_options.namespace}/"
                + f"{model_spec.model_details.name}"
                + model_spec.protocol.get_predict_path(model_spec.model_details)
            )
        else:
            # TODO check why needed this here
            config.load_incluster_config()
            api_instance = client.CustomObjectsApi()
            api_response = api_instance.get_namespaced_custom_object_status(
                "machinelearning.seldon.io",
                "v1",
                model_spec.runtime_options.k8s_options.namespace,
                "seldondeployments",
                model_spec.model_details.name,
            )
            return api_response["status"]["address"]["url"]



from ast import literal_eval
from typing import Any, Dict, List, Optional, Type, Union

import numpy as np

from tempo.serve.metadata import ModelDataArgs, ModelDetails
from tempo.serve.protocol import Protocol

_v2tymap: Dict[str, np.dtype] = {
    "BOOL": np.dtype("bool"),
    "UINT8": np.dtype("uint8"),
    "UINT16": np.dtype("uint16"),
    "UINT32": np.dtype("uint32"),
    "UINT64": np.dtype("uint64"),
    "INT8": np.dtype("int8"),
    "INT16": np.dtype("int16"),
    "INT32": np.dtype("int32"),
    "INT64": np.dtype("int64"),
    "FP16": np.dtype("float32"),
    "FP32": np.dtype("float32"),
    "FP64": np.dtype("float64"),
}

_nptymap = dict([(value, key) for key, value in _v2tymap.items()])
_nptymap[np.dtype("float32")] = "FP32"  # Ensure correct mapping for ambiguous type


class KFServingV2Protocol(Protocol):
    @staticmethod
    def create_v2_from_np(arr: np.ndarray, name: str) -> Dict:
        if arr.dtype in _nptymap:
            return {
                "name": name,
                "datatype": _nptymap[arr.dtype],
                "data": arr.flatten().tolist(),
                "shape": list(arr.shape),
            }
        else:
            raise ValueError(f"Unknown numpy type {arr.dtype}")

    @staticmethod
    def create_v2_from_any(data: Any, name: str) -> Dict:
        if isinstance(data, str):
            b = list(bytes(data, "utf-8"))
        else:
            b = list(bytes(repr(data), "utf-8"))
        return {
            "name": name,
            "datatype": "BYTES",
            "data": b,
            "shape": [len(b)],
        }

    @staticmethod
    def convert_from_bytes(output: dict, ty: Optional[Type]) -> Any:
        if ty == str:
            return bytearray(output["data"]).decode("UTF-8")
        else:
            py_str = bytearray(output["data"]).decode("UTF-8")
            return literal_eval(py_str)

    @staticmethod
    def create_np_from_v2(data: list, ty: str, shape: list) -> np.ndarray:
        if ty in _v2tymap:
            npty = _v2tymap[ty]
            arr = np.array(data, dtype=npty)
            arr.shape = tuple(shape)
            return arr
        else:
            raise ValueError(f"V2 unknown type or type that can't be coerced {ty}")

    def get_predict_path(self, model_details: ModelDetails):
        return f"/v2/models/{model_details.name}/infer"

    def get_status_path(self, model_details: ModelDetails) -> str:
        return f"/v2/models/{model_details.name}/ready"

    def to_protocol_request(self, *args, **kwargs) -> Dict:
        # if len(args) > 0:
        #    raise ValueError("KFserving V2 protocol only supports named arguments")

        inputs = []
        if len(args) > 0:
            for idx, raw in enumerate(args):
                raw_type = type(raw)
                if raw_type == np.ndarray:
                    inputs.append(KFServingV2Protocol.create_v2_from_np(raw, "input-" + str(idx)))
                else:
                    inputs.append(KFServingV2Protocol.create_v2_from_any(raw, "input-" + str(idx)))
        else:
            for (name, raw) in kwargs.items():
                raw_type = type(raw)

                if raw_type == np.ndarray:
                    inputs.append(KFServingV2Protocol.create_v2_from_np(raw, name))
                else:
                    inputs.append(KFServingV2Protocol.create_v2_from_any(raw, name))

        return {"inputs": inputs}

    @staticmethod
    def get_ty(name: str, idx: int, tys: ModelDataArgs) -> Optional[Type]:
        ty = tys[name]
        if ty is None:
            ty = tys[idx]
        # if ty is None:
        #    return np.ndarray
        return ty

    def to_protocol_response(self, model_details: ModelDetails, *args, **kwargs) -> Dict:
        outputs = []
        for idx, raw in enumerate(args):
            raw_type = type(raw)

            if raw_type == np.ndarray:
                outputs.append(KFServingV2Protocol.create_v2_from_np(raw, "output" + str(idx)))
            else:
                outputs.append(KFServingV2Protocol.create_v2_from_any(raw, "output" + str(idx)))
        for name, raw in kwargs.items():
            raw_type = type(raw)

            if raw_type == np.ndarray:
                shape = list(raw.shape)
                data = raw.flatten().tolist()
                outputs.append({"name": name, "datatype": "FP32", "shape": shape, "data": data})
            else:
                outputs.append(KFServingV2Protocol.create_v2_from_any(raw, name))
        return {"model_name": model_details.name, "outputs": outputs}

    def from_protocol_request(self, res: Dict, tys: ModelDataArgs) -> Any:
        inp = {}
        for idx, input in enumerate(res["inputs"]):
            ty = KFServingV2Protocol.get_ty(input["name"], idx, tys)

            if ty == np.ndarray:
                arr = KFServingV2Protocol.create_np_from_v2(input["data"], input["datatype"], input["shape"])
                inp[input["name"]] = arr
            elif input["datatype"] == "BYTES":
                inp[input["name"]] = KFServingV2Protocol.convert_from_bytes(input, ty)
            else:
                raise ValueError(f"Unknown ty {ty} in conversion")

        if len(inp) == 1:
            return list(inp.values())[0]
        else:
            return inp

    def from_protocol_response(self, res: Dict, tys: ModelDataArgs) -> Any:
        out = {}
        for idx, output in enumerate(res["outputs"]):
            ty = KFServingV2Protocol.get_ty(output["name"], idx, tys)

            if ty == np.ndarray:
                arr = KFServingV2Protocol.create_np_from_v2(output["data"], output["datatype"], output["shape"])
                out[output["name"]] = arr
            elif output["datatype"] == "BYTES":
                out[output["name"]] = KFServingV2Protocol.convert_from_bytes(output, ty)
            else:
                raise ValueError(f"Unknown ty {ty} in conversion")

        if len(out) == 1:
            return list(out.values())[0]
        else:
            return out


class KFServingV1Protocol(Protocol):
    @staticmethod
    def create_v1_from_np(arr: np.ndarray, name: str = None) -> list:
        return arr.tolist()

    @staticmethod
    def create_np_from_v1(data: list) -> np.ndarray:
        arr = np.array(data)
        return arr

    def get_predict_path(self, model_details: ModelDetails):
        return f"/v1/models/{model_details.name}:predict"

    def get_status_path(self, model_details: ModelDetails) -> str:
        return f"/v1/models/{model_details.name}"

    def to_protocol_request(self, *args, **kwargs) -> Dict:
        if len(args) > 0 and len(kwargs.values()) > 0:
            raise ValueError("KFserving V1 protocol only supports either named or unamed arguments but not both")

        inputs = []
        if len(args) > 0:
            for raw in args:
                raw_type = type(raw)

                if raw_type == np.ndarray:
                    inputs.append(KFServingV1Protocol.create_v1_from_np(raw))
        else:
            for (name, raw) in kwargs.items():
                raw_type = type(raw)

                if raw_type == np.ndarray:
                    inputs.append(KFServingV1Protocol.create_v1_from_np(raw, name))
                else:
                    raise ValueError(f"Unknown input type {raw_type}")

        if len(inputs) == 1:
            return {"instances": inputs[0]}
        else:
            return {"instances": inputs}

    @staticmethod
    def get_ty(name: Optional[str], idx: int, tys: ModelDataArgs) -> Type:
        ty = None
        if name is not None:
            ty = tys[name]
        if ty is None:
            ty = tys[idx]
        if ty is None:
            return np.ndarray
        return ty

    def to_protocol_response(self, model_details: ModelDetails, *args, **kwargs) -> Dict:
        outputs: List[Union[Dict, List]] = []
        if len(args) > 0:
            for idx, raw in enumerate(args):
                raw_type = type(raw)

                if raw_type == np.ndarray:
                    outputs.append(KFServingV1Protocol.create_v1_from_np(raw))
                else:
                    raise ValueError(f"Unknown input type {raw_type}")
        else:
            for name, raw in kwargs.items():
                raw_type = type(raw)

                if raw_type == np.ndarray:
                    data = raw.tolist()
                    outputs.append({name: data})
                else:
                    raise ValueError(f"Unknown input type {raw_type}")
        return {"predictions": outputs}

    def from_protocol_request(self, res: Dict, tys: ModelDataArgs) -> Any:
        inp = {}
        for idx, input in enumerate(res["inputs"]):
            ty = KFServingV1Protocol.get_ty(input["name"], idx, tys)

            if ty == np.ndarray:
                arr = KFServingV2Protocol.create_np_from_v2(input["data"], input["datatype"], input["shape"])
                inp[input["name"]] = arr
            else:
                raise ValueError(f"Unknown ty {ty} in conversion")

        if len(inp) == 1:
            return list(inp.values())[0]
        else:
            return inp

    def from_protocol_response(self, res: Dict, tys: ModelDataArgs) -> Any:
        if len(tys) <= 1:
            ty = KFServingV1Protocol.get_ty(None, 0, tys)

            if ty == np.ndarray:
                return KFServingV1Protocol.create_np_from_v1(res["predictions"])
            else:
                raise ValueError(f"Unknown ty {ty} in conversion")
        else:
            out = []
            for idx, output in enumerate(res["predictions"]):
                if type(output) == list:
                    for idx2, it in enumerate(output):
                        ty = KFServingV1Protocol.get_ty(None, idx, tys)

                        if ty == np.ndarray:
                            arr = KFServingV1Protocol.create_np_from_v1(it)
                            out.append(arr)
                        else:
                            raise ValueError(f"Unknown ty {ty} in conversion")

            if len(out) == 1:
                return out[0]
            else:
                return out



import numpy as np
import pytest

from tempo.kfserving.k8s import KFServingKubernetesRuntime
from tempo.serve.metadata import ModelFramework
from tempo.serve.model import Model


@pytest.mark.skip(reason="needs k8s cluster")
def test_kfserving():
    model = Model(
        name="sklearn-iris2",
        runtime=KFServingKubernetesRuntime(),
        platform=ModelFramework.SKLearn,
        uri="gs://kfserving-samples/models/sklearn/iris",
        local_folder="sklearn/model",
    )
    # model.deploy()
    res = model(np.array([[1, 2, 3, 4]]))
    print(res)



import numpy as np
import pytest


@pytest.mark.parametrize(
    "x_input, expected",
    [(np.array([[0.4, 2, 3, 4]]), [9.4])],
)
def test_class2(inference_pipeline_class, x_input, expected):
    y_pred = inference_pipeline_class.pipeline(x_input)
    assert y_pred == expected
    y_pred = inference_pipeline_class.pipeline(x_input)
    assert y_pred == expected
    assert inference_pipeline_class.get_counter() == 2



import os

import pytest

from tempo.serve.constants import DefaultEnvFilename, MLServerEnvDeps
from tempo.serve.loader import (
    _add_required_deps,
    _create_and_pack_environment,
    _get_env,
    _get_environment,
    _get_pip_deps,
    _has_required_deps,
    save_environment,
)


@pytest.mark.skip(reason="very slow")
def test_save_environment(tmp_path):
    env_path = os.path.join(tmp_path, DefaultEnvFilename)
    save_environment(conda_pack_file_path=env_path, env_name="base")

    assert os.path.isfile(env_path)


def test_create_and_pack_environment(tmp_path):
    env = {"dependencies": [{"pip": MLServerEnvDeps}]}
    env_path = os.path.join(tmp_path, DefaultEnvFilename)

    _create_and_pack_environment(env=env, file_path=env_path)

    assert os.path.isfile(env_path)


def test_get_environment():
    env = _get_environment(env_name="base")

    assert env["name"] == "base"


@pytest.mark.parametrize(
    "env",
    [
        {},
        {"foo": "bar"},
        {"foo": "bar", "dependencies": ["foo==1.0"]},
        {
            "foo": "bar",
            "dependencies": ["foo==1.0", {"pip": ["foo-pip==1.0"]}],
        },
    ],
)
def test_add_required_deps(env):
    env_with_deps = _add_required_deps(env=env)

    pip_deps = _get_pip_deps(env_with_deps["dependencies"])
    for dep in MLServerEnvDeps:
        assert dep in pip_deps["pip"]


@pytest.mark.parametrize(
    "env, expected",
    [
        ({}, False),
        (
            {
                "foo": "bar",
                "dependencies": ["foo==1.0", {"pip": ["mlserver"]}],
            },
            True,
        ),
    ],
)
def test_has_required_deps(env, expected):
    has_deps = _has_required_deps(env)
    assert has_deps == expected


@pytest.mark.parametrize(
    "env_file_path",
    [
        os.path.join(os.path.dirname(__file__), "data", "conda.yaml"),
    ],
)
def test_get_env_ok(env_file_path):
    _get_env(conda_env_file_path=env_file_path)


@pytest.mark.parametrize(
    "env_file_path",
    [
        os.path.join(os.path.dirname(__file__), "data", "conda_missing_mlserver.yaml"),
    ],
)
def test_get_env_not_ok(env_file_path):
    with pytest.raises(ValueError):
        _get_env(conda_env_file_path=env_file_path)



import pytest
from mlserver.settings import ModelParameters, ModelSettings
from mlserver.types import InferenceRequest, RequestInput
from mlserver.utils import to_ndarray

from tempo import Model
from tempo.mlserver import InferenceRuntime
from tempo.serve.loader import save


@pytest.fixture
def model_settings(custom_model: Model) -> ModelSettings:
    save(custom_model, save_env=False)
    pipeline_uri = custom_model.details.local_folder

    return ModelSettings(
        name="sum-model",
        parameters=ModelParameters(uri=pipeline_uri),
    )


@pytest.fixture
def inference_request() -> InferenceRequest:
    return InferenceRequest(inputs=[RequestInput(name="input-0", shape=[4], data=[1, 2, 3, 4], datatype="FP32")])


@pytest.fixture
async def model(model_settings: ModelSettings) -> InferenceRuntime:
    model = InferenceRuntime(model_settings)
    await model.load()

    return model


def test_load(model: InferenceRuntime):
    assert model.ready
    assert isinstance(model._pipeline, Model)


async def test_predict(
    model: InferenceRuntime,
    inference_request: InferenceRequest,
    custom_model: Model,
):
    res = await model.predict(inference_request)

    assert len(res.outputs) == 1

    pipeline_input = to_ndarray(inference_request.inputs[0])
    custom_model.get_tempo().use_remote = False  # ensure direct call to class does not try to do remote
    expected_output = custom_model(pipeline_input)

    pipeline_output = res.outputs[0].data

    assert expected_output.tolist() == pipeline_output



import numpy as np
import pytest

from tempo.seldon.deploy import SeldonDeployAuthType, SeldonDeployConfig, SeldonDeployRuntime
from tempo.seldon.k8s import SeldonKubernetesRuntime
from tempo.seldon.protocol import SeldonProtocol
from tempo.serve.metadata import IngressOptions, KubernetesOptions, ModelFramework, RuntimeOptions
from tempo.serve.model import Model


@pytest.mark.skip("needs deploy cluster")
def test_deploy():
    rt = SeldonDeployRuntime()

    config = SeldonDeployConfig(
        host="https://34.78.44.92/seldon-deploy/api/v1alpha1",
        user="admin@seldon.io",
        password="12341234",
        oidc_server="https://34.78.44.92/auth/realms/deploy-realm",
        oidc_client_id="sd-api",
        verify_ssl=False,
        auth_type=SeldonDeployAuthType.oidc,
    )

    rt.authenticate(settings=config)

    options = RuntimeOptions(
        runtime="tempo.seldon.SeldonKubernetesRuntime",
        k8s_options=KubernetesOptions(namespace="seldon"),
        ingress_options=IngressOptions(ssl=True, verify_ssl=False),
    )

    sklearn_model = Model(
        name="test-iris-sklearn",
        platform=ModelFramework.SKLearn,
        uri="gs://seldon-models/sklearn/iris",
        protocol=SeldonProtocol(),
        runtime_options=options,
    )

    rt.deploy(sklearn_model)
    rt.wait_ready(sklearn_model)
    print(sklearn_model(np.array([[4.9, 3.1, 1.5, 0.2]])))
    rt.undeploy(sklearn_model)


@pytest.mark.skip("needs deploy cluster")
def test_deploy_yaml():
    rt = SeldonDeployRuntime(
        host="http://34.78.44.92/seldon-deploy/api/v1alpha1",
        user="admin@kubeflow.org",
        oidc_server="https://34.78.44.92/auth/realms/deploy-realm",
        password="12341234",
        oidc_client_id="sd-api",
        verify_ssl=False,
    )

    options = RuntimeOptions(
        runtime="tempo.seldon.SeldonKubernetesRuntime",
        k8s_options=KubernetesOptions(namespace="seldon"),
        ingress_options=IngressOptions(ssl=True, verify_ssl=False),
    )

    sklearn_model = Model(
        name="test-iris-sklearn",
        platform=ModelFramework.SKLearn,
        uri="gs://seldon-models/sklearn/iris",
        protocol=SeldonProtocol(),
        runtime_options=options,
    )

    spec = rt.to_k8s_yaml(sklearn_model)
    rtk = SeldonKubernetesRuntime()
    expected = rtk.to_k8s_yaml(sklearn_model)
    assert spec == expected



from pathlib import Path

from pydantic import BaseSettings


class TempoSettings(BaseSettings):
    rclone_cfg: str = str(Path.home()) + "/.config/rclone/rclone.conf"
    use_kubernetes: bool = False


settings = TempoSettings()



from __future__ import annotations

import abc
from typing import Any, Optional

from pydantic import BaseModel

from tempo.serve.metadata import ModelDetails, RuntimeOptions
from tempo.serve.protocol import Protocol


class ModelSpec(BaseModel):

    model_details: ModelDetails
    protocol: Protocol
    runtime_options: RuntimeOptions

    class Config:
        arbitrary_types_allowed = True


class Deployer(object):
    def __init__(self, runtime_options: Optional[RuntimeOptions]):
        self.runtime_options = runtime_options

    def deploy(self, model: Any):
        t = model.get_tempo()
        t.set_runtime_options_override(self.runtime_options)
        t.deploy(self)

    def undeploy(self, model: Any):
        t = model.get_tempo()
        t.set_runtime_options_override(self.runtime_options)
        t.undeploy(self)

    def get_endpoint(self, model: Any):
        t = model.get_tempo()
        t.set_runtime_options_override(self.runtime_options)
        t.get_endpoint(self)

    def wait_ready(self, model: Any, timeout_secs=None):
        t = model.get_tempo()
        t.set_runtime_options_override(self.runtime_options)
        t.wait_ready(self, timeout_secs)

    def to_k8s_yaml(self, model: Any):
        t = model.get_tempo()
        t.set_runtime_options_override(self.runtime_options)
        return t.to_k8s_yaml(self)


class Runtime(abc.ABC, Deployer):
    # TODO change to deploy_model
    @abc.abstractmethod
    def deploy_spec(self, model_spec: ModelSpec):
        pass

    # TODO change to undeploy_model
    @abc.abstractmethod
    def undeploy_spec(self, model_spec: ModelSpec):
        pass

    @abc.abstractmethod
    def get_endpoint_spec(self, model_spec: ModelSpec) -> str:
        pass

    @abc.abstractmethod
    def wait_ready_spec(self, model_spec: ModelSpec, timeout_secs=None) -> bool:
        pass

    # TODO change to to_yaml
    @abc.abstractmethod
    def to_k8s_yaml_spec(self, model_spec: ModelSpec) -> str:
        pass


class LocalRuntime(Runtime):
    """
    LocalRuntime lets you run model functions explicitly as a local function.
    """

    def __init__(self):
        super().__init__()

    def deploy(self, model_details: ModelDetails):
        pass

    def undeploy(self, model_spec: ModelDetails):
        pass

    def remote(self, model_spec: ModelDetails, *args, **kwargs) -> Any:
        raise NotImplementedError("LocalRuntime is only meant to be used locally")

    def get_endpoint(self, model_spec: ModelDetails) -> str:
        return ""

    def wait_ready(self, model_spec: ModelDetails, timeout_secs=None) -> bool:
        return True

    def to_k8s_yaml(self, model_spec: ModelDetails) -> str:
        return ""



from kubernetes import client, config

from tempo.serve.ingress import Ingress
from tempo.serve.runtime import ModelSpec


class IstioIngress(Ingress):
    def get_external_host_url(self, model_spec: ModelSpec) -> str:
        config.load_kube_config()
        api_instance = client.CoreV1Api()
        res = api_instance.list_namespaced_service("istio-system", field_selector="metadata.name=istio-ingressgateway")
        ingress_ip = res.items[0].status.load_balancer.ingress[0].ip
        if not ingress_ip:
            ingress_ip = res.items[0].status.load_balancer.ingress[0].hostname
        scheme = "http"
        if model_spec.runtime_options.ingress_options.ssl:
            scheme = "https"
        return f"{scheme}://{ingress_ip}"



from typing import Any, Dict, Type, Union

import numpy as np

from tempo.serve.metadata import ModelDataArgs, ModelDetails
from tempo.serve.protocol import Protocol


class SeldonProtocol(Protocol):
    def get_predict_path(self, model_details: ModelDetails):
        # TODO: K8s backend needs to add namespace and model name
        #  return f"/seldon/{model_details.namespace}/{model_details.name}/api/v1.0/predictions"
        return "/api/v1.0/predictions"

    def get_status_path(self, model_details: ModelDetails) -> str:
        return "/api/v1.0/health/status"

    def to_protocol_request(self, *args, **kwargs) -> Dict:
        if not len(args) + len(kwargs) == 1:
            raise ValueError("Seldon protocol can only take a single input")

        if len(args) == 1:
            raw = args[0]
        else:
            raw = list(kwargs.values())[0]

        raw_type = type(raw)

        if raw_type == dict:
            return raw
        elif raw_type == list:
            return {"data": {"ndarray": raw}}
        elif raw_type == np.ndarray:
            return {"data": {"ndarray": raw.tolist()}}

        raise ValueError(f"Unknown input type {raw_type}")

    def to_protocol_response(self, model_details: ModelDetails, *args, **kwargs) -> Dict:
        return self.to_protocol_request(*args, **kwargs)

    def from_protocol_request(self, res: dict, tys: ModelDataArgs) -> Union[dict, np.ndarray]:
        if len(tys) > 1:
            raise ValueError("Seldon protocol can only return a single type")

        if tys[0] == Dict:
            # Return as-is
            return res

        ty: Type = np.ndarray
        if tys and tys[0] is not None:
            ty = tys[0]

        if ty == np.ndarray:
            if "data" in res:
                datadef = res["data"]
                if "tensor" in datadef:
                    tensor = datadef["tensor"]
                    return np.array(tensor["values"]).reshape(tensor["shape"])
                elif "ndarray" in datadef:
                    return np.array(datadef["ndarray"])
                # elif "tftensor" in datadef:
                #    tf_proto = TensorProto()
                #    json_format.ParseDict(datadef["tftensor"], tf_proto)
                #    return tf.make_ndarray(tf_proto)

            raise ValueError(f"Unknown data structure {res}")

        raise ValueError(f"unknown type {ty}")

    def from_protocol_response(self, res: Dict, tys: ModelDataArgs) -> Any:
        return self.from_protocol_request(res, tys)



from .serve.metadata import ModelFramework
from .serve.model import Model
from .serve.pipeline import Pipeline
from .serve.utils import model, pipeline, predictmethod
from .version import __version__

__all__ = [
    "__version__",
    "ModelFramework",
    "Model",
    "Pipeline",
    "pipeline",
    "predictmethod",
    "model",
]



class TempoError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class UndefinedRuntime(TempoError):
    def __init__(self, model_name: str):
        msg = f"Undefined runtime for model {model_name}"
        super().__init__(msg)


class UndefinedCustomImplementation(TempoError):
    def __init__(self, model_name: str):
        msg = f"Undefined custom implementation for model {model_name}"
        super().__init__(msg)



import os
from typing import Dict

from setuptools import find_packages, setup

ROOT_PATH = os.path.dirname(__file__)
PKG_NAME = "tempo"
PKG_PATH = os.path.join(ROOT_PATH, PKG_NAME)


def _load_version() -> str:
    version = ""
    version_path = os.path.join(PKG_PATH, "version.py")
    with open(version_path) as fp:
        version_module: Dict[str, str] = {}
        exec(fp.read(), version_module)
        version = version_module["__version__"]

    return version


def _load_description() -> str:
    readme_path = os.path.join(ROOT_PATH, "README.md")
    with open(readme_path) as fp:
        return fp.read()


setup(
    #  name=PKG_NAME,
    # TODO: Update once we've got consensus on package name
    name="mlops-tempo",
    author="Seldon Technologies Ltd.",
    author_email="hello@seldon.io",
    version=_load_version(),
    description="Machine Learning Operations Toolkit",
    url="https://github.com/SeldonIO/tempo",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6",
    setup_requires=["pytest-runner"],
    install_requires=[
        "grpcio>=1.32.0",
        "protobuf>=3.14.0",
        "attrs",
        "numpy",
        "kubernetes",
        "docker",
        "packaging",
        "requests",
        "pydantic",
        "cloudpickle",
        "python-rclone",
        "seldon-deploy-sdk",
        "conda-pack",
        "mlserver",
    ],
    tests_require=["pytest", "pytest-cov", "pytest-xdist", "pytest-lazy-fixture"],
    zip_safe=False,
    long_description=_load_description(),
    long_description_content_type="text/markdown",
    license="Apache 2.0",
)



from __future__ import annotations

import os
import tempfile
from os import path
from pydoc import locate
from typing import Any, Callable, Dict, Optional, Tuple, get_type_hints

import numpy as np

from tempo.conf import settings
from tempo.errors import UndefinedCustomImplementation
from tempo.kfserving.protocol import KFServingV2Protocol
from tempo.serve.constants import (
    ENV_K8S_SERVICE_HOST,
    DefaultCondaFile,
    DefaultEnvFilename,
    DefaultModelFilename,
    ModelDataType,
)
from tempo.serve.loader import load_custom, save_custom, save_environment
from tempo.serve.metadata import ModelDataArg, ModelDataArgs, ModelDetails, ModelFramework, RuntimeOptions
from tempo.serve.protocol import Protocol
from tempo.serve.remote import Remote
from tempo.serve.runtime import ModelSpec, Runtime
from tempo.utils import logger


class BaseModel:
    def __init__(
        self,
        name: str,
        user_func: Callable[..., Any] = None,
        local_folder: str = None,
        uri: str = None,
        platform: ModelFramework = None,
        inputs: ModelDataType = None,
        outputs: ModelDataType = None,
        conda_env: str = None,
        protocol: Protocol = None,
        runtime_options: RuntimeOptions = RuntimeOptions(),
    ):
        if protocol is None:
            protocol = KFServingV2Protocol()
        self._name = name
        self._user_func = user_func
        self.conda_env_name = conda_env

        if uri is None:
            uri = ""

        local_folder = self._get_local_folder(local_folder)
        input_args, output_args = self._get_args(inputs, outputs)

        self.details = ModelDetails(
            name=name,
            local_folder=local_folder,
            uri=uri,
            platform=platform,
            inputs=input_args,
            outputs=output_args,
        )

        self.cls = None
        self.protocol = protocol
        self.model_spec = ModelSpec(model_details=self.details, protocol=self.protocol, runtime_options=runtime_options)

        self.use_remote: bool = False
        self.runtime_options_override: Optional[RuntimeOptions] = None

    def set_remote(self, val: bool):
        self.use_remote = val

    def set_runtime_options_override(self, runtime_options: RuntimeOptions):
        self.runtime_options_override = runtime_options

    def _get_args(
        self, inputs: ModelDataType = None, outputs: ModelDataType = None
    ) -> Tuple[ModelDataArgs, ModelDataArgs]:
        input_args = []
        output_args = []

        if isinstance(inputs, ModelDataArgs) and isinstance(outputs, ModelDataArgs):
            return inputs, outputs
        elif inputs is None and outputs is None:
            if self._user_func is not None:
                hints = get_type_hints(self._user_func)
                for k, v in hints.items():
                    if k == "return":
                        if hasattr(v, "__args__"):
                            # NOTE: If `__args__` are present, assume this as a
                            # `typing.Generic`, like `Tuple`
                            targs = v.__args__
                            for targ in targs:
                                output_args.append(ModelDataArg(ty=targ))
                        else:
                            output_args.append(ModelDataArg(ty=v))
                    else:
                        input_args.append(ModelDataArg(name=k, ty=v))
            else:
                input_args.append(ModelDataArg(ty=np.ndarray))
                output_args.append(ModelDataArg(ty=np.ndarray))
        else:
            if isinstance(outputs, dict):
                for k, v in outputs.items():
                    output_args.append(ModelDataArg(name=k, ty=v))
            elif isinstance(outputs, tuple):
                for ty in list(outputs):
                    output_args.append(ModelDataArg(ty=ty))
            else:
                output_args.append(ModelDataArg(ty=outputs))

            if isinstance(inputs, dict):
                for k, v in inputs.items():
                    input_args.append(ModelDataArg(name=k, ty=v))
            elif isinstance(inputs, tuple):
                for ty in list(inputs):
                    input_args.append(ModelDataArg(ty=ty))
            else:
                input_args.append(ModelDataArg(ty=inputs))

        return ModelDataArgs(args=input_args), ModelDataArgs(args=output_args)

    def _get_local_folder(self, local_folder: str = None) -> Optional[str]:
        if not local_folder:
            # TODO: Should we do cleanup at some point?
            local_folder = tempfile.mkdtemp()

        return local_folder

    def set_cls(self, cls):
        self.cls = cls

    @classmethod
    def load(cls, folder: str) -> "BaseModel":
        file_path_pkl = os.path.join(folder, DefaultModelFilename)
        return load_custom(file_path_pkl)

    def save(self, save_env=True):
        logger.info("Saving environment")
        if not self._user_func:
            # Nothing to save
            return

        file_path_pkl = os.path.join(self.details.local_folder, DefaultModelFilename)
        logger.info("Saving tempo model to %s", file_path_pkl)
        save_custom(self, file_path_pkl)

        if save_env:
            file_path_env = os.path.join(self.details.local_folder, DefaultEnvFilename)
            conda_env_file_path = path.join(self.details.local_folder, DefaultCondaFile)
            if not path.exists(conda_env_file_path):
                conda_env_file_path = None

            save_environment(
                conda_pack_file_path=file_path_env,
                conda_env_file_path=conda_env_file_path,
                env_name=self.conda_env_name,
            )

    def request(self, req: Dict) -> Dict:

        if self._user_func is None:
            raise UndefinedCustomImplementation(self.details.name)

        req_converted = self.protocol.from_protocol_request(req, self.details.inputs)
        if type(req_converted) == dict:
            if self.cls is not None:
                response = self._user_func(self.cls, **req_converted)
            else:
                response = self._user_func(**req_converted)
        elif type(req_converted) == list or type(req_converted) == tuple:
            if self.cls is not None:
                response = self._user_func(self.cls, *req_converted)
            else:
                response = self._user_func(*req_converted)
        else:
            if self.cls is not None:
                response = self._user_func(self.cls, req_converted)
            else:
                response = self._user_func(req_converted)

        if type(response) == dict:
            response_converted = self.protocol.to_protocol_response(self.details, **response)
        elif type(response) == list or type(response) == tuple:
            response_converted = self.protocol.to_protocol_response(self.details, *response)
        else:
            response_converted = self.protocol.to_protocol_response(self.details, response)

        return response_converted

    def _get_model_spec(self) -> ModelSpec:
        if self.runtime_options_override:
            return ModelSpec(
                model_details=self.details, protocol=self.protocol, runtime_options=self.runtime_options_override
            )
        else:
            return self.model_spec

    def _create_remote(self, model_spec: ModelSpec) -> Remote:
        cls_path = model_spec.runtime_options.runtime
        if cls_path is None:
            if settings.use_kubernetes or os.getenv(ENV_K8S_SERVICE_HOST):
                cls_path = model_spec.runtime_options.k8s_options.defaultRuntime
            else:
                cls_path = model_spec.runtime_options.docker_options.defaultRuntime
        logger.debug("Using remote class %s", cls_path)
        cls: Any = locate(cls_path)
        return cls()

    def remote(self, *args, **kwargs):
        remoter = self._create_remote(self._get_model_spec())
        return remoter.remote(self._get_model_spec(), *args, **kwargs)

    def wait_ready(self, runtime: Runtime, timeout_secs=None):
        return runtime.wait_ready_spec(self._get_model_spec(), timeout_secs=timeout_secs)

    def get_endpoint(self, runtime: Runtime):
        return runtime.get_endpoint_spec(self._get_model_spec())

    def to_k8s_yaml(self, runtime: Runtime) -> str:
        """
        Get k8s yaml
        """

        return runtime.to_k8s_yaml_spec(self._get_model_spec())

    def deploy(self, runtime: Runtime):
        # self.set_runtime(runtime)
        runtime.deploy_spec(self._get_model_spec())

    def undeploy(self, runtime: Runtime):
        # self.unset_runtime()
        logger.info("Undeploying %s", self.details.name)
        runtime.undeploy_spec(self._get_model_spec())

    def get_tempo(self) -> BaseModel:
        return self

    def __call__(self, *args, **kwargs) -> Any:
        if not self._user_func:
            return self.remote(*args, **kwargs)
        else:
            if self.use_remote:
                return self.remote(*args, **kwargs)
            else:
                if self.cls is not None:
                    return self._user_func(self.cls, *args, **kwargs)
                else:
                    return self._user_func(*args, **kwargs)



import time
from typing import Generator

import docker
import pytest

from tempo import Model, Pipeline
from tempo.seldon import SeldonDockerRuntime
from tempo.serve.loader import save
from tempo.serve.metadata import RuntimeOptions


@pytest.fixture
def runtime() -> SeldonDockerRuntime:
    return SeldonDockerRuntime(RuntimeOptions())


@pytest.fixture
def sklearn_model(sklearn_model: Model, runtime: SeldonDockerRuntime) -> Generator[Model, None, None]:
    runtime.deploy(sklearn_model)
    runtime.wait_ready(sklearn_model, timeout_secs=60)

    yield sklearn_model

    try:
        runtime.undeploy(sklearn_model)
    except docker.errors.NotFound:
        # TODO: Should undeploy be idempotent as well?
        # Ignore if the model has already been undeployed
        pass


@pytest.fixture
def xgboost_model(xgboost_model: Model, runtime: SeldonDockerRuntime) -> Generator[Model, None, None]:
    runtime.deploy(xgboost_model)
    runtime.wait_ready(xgboost_model, timeout_secs=60)

    yield xgboost_model

    try:
        runtime.undeploy(xgboost_model)
    except docker.errors.NotFound:
        # TODO: Should undeploy be idempotent as well?
        # Ignore if the model has already been undeployed
        pass


@pytest.fixture
def cifar10_model(cifar10_model: Model, runtime: SeldonDockerRuntime) -> Generator[Model, None, None]:
    runtime.deploy(cifar10_model)
    runtime.wait_ready(cifar10_model, timeout_secs=60)

    yield cifar10_model

    runtime.undeploy(cifar10_model)


@pytest.fixture
def inference_pipeline(
    inference_pipeline: Pipeline,
    runtime: SeldonDockerRuntime,
) -> Generator[Pipeline, None, None]:

    # NOTE: Need to re-save the pipeline so that it knows about the runtime
    save(inference_pipeline, save_env=True)
    runtime.deploy(inference_pipeline)
    runtime.wait_ready(inference_pipeline, timeout_secs=60)
    # TODO: Fix wait_ready for pipelines
    time.sleep(8)

    yield inference_pipeline

    try:
        runtime.undeploy(inference_pipeline)
    except docker.errors.NotFound:
        # TODO: Should undeploy be idempotent as well?
        # Ignore if the model has already been undeployed
        pass
